"""This module encodes and decodes various types of base58 encodings."""

from typing import Dict, List, Optional, Tuple

import base58
from typing_extensions import Final

from xrpl.constants import CryptoAlgorithm
from xrpl.core.addresscodec.exceptions import XRPLAddressCodecException
from xrpl.core.addresscodec.utils import XRPL_ALPHABET

# base58 encodings: https://xrpl.org/base58-encodings.html
# Account address (20 bytes)
_CLASSIC_ADDRESS_PREFIX: Final[List[int]] = [0x0]
# value is 35; Account public key (33 bytes)
_ACCOUNT_PUBLIC_KEY_PREFIX: Final[List[int]] = [0x23]
# value is 33; Seed value (for secret keys) (16 bytes)
_FAMILY_SEED_PREFIX: Final[List[int]] = [0x21]
# value is 28; Validation public key (33 bytes)
_NODE_PUBLIC_KEY_PREFIX: Final[List[int]] = [0x1C]
# [1, 225, 75]
_ED25519_SEED_PREFIX: Final[List[int]] = [0x01, 0xE1, 0x4B]

SEED_LENGTH: Final[int] = 16

_CLASSIC_ADDRESS_LENGTH: Final[int] = 20
_NODE_PUBLIC_KEY_LENGTH: Final[int] = 33
_ACCOUNT_PUBLIC_KEY_LENGTH: Final[int] = 33

_ALGORITHM_TO_PREFIX_MAP: Final[Dict[CryptoAlgorithm, List[List[int]]]] = {
    CryptoAlgorithm.ED25519: [_ED25519_SEED_PREFIX, _FAMILY_SEED_PREFIX],
    CryptoAlgorithm.SECP256K1: [_FAMILY_SEED_PREFIX],
}  # first is default, rest are other options


def _encode(bytestring: bytes, prefix: List[int], expected_length: int) -> str:
    """
    Returns the base58 encoding of the bytestring, with the given data prefix
    (which indicates type) and while ensuring the bytestring is the expected
    length.
    """
    if expected_length and len(bytestring) != expected_length:
        error_message = """unexpected_payload_length: len(bytestring) does not match expected_length.
        Ensure that the bytes are a bytestring."""
        raise XRPLAddressCodecException(error_message)
    encoded_prefix = bytes(prefix)
    payload = encoded_prefix + bytestring
    return base58.b58encode_check(payload, alphabet=XRPL_ALPHABET).decode("utf-8")


def _decode(b58_string: str, prefix: bytes) -> bytes:
    """
    Args:
        b58_string: A base58 value.
        prefix: The prefix prepended to the bytestring.

    Returns:
        The byte decoding of the base58-encoded string.
    """
    prefix_length = len(prefix)
    decoded = base58.b58decode_check(b58_string, alphabet=XRPL_ALPHABET)
    if decoded[:prefix_length] != prefix:
        raise XRPLAddressCodecException("Provided prefix is incorrect")
    return decoded[prefix_length:]


def encode_seed(entropy: bytes, encoding_type: CryptoAlgorithm) -> str:
    """
    Returns an encoded seed.

    Args:
        entropy: Entropy bytes of SEED_LENGTH.
        encoding_type: Either ED25519 or SECP256K1.

    Returns:
        An encoded seed.

    Raises:
        XRPLAddressCodecException: If entropy is not of length SEED_LENGTH
            or the encoding type is not one of CryptoAlgorithm.
    """
    if len(entropy) != SEED_LENGTH:
        raise XRPLAddressCodecException(f"Entropy must have length {SEED_LENGTH}")
    if encoding_type not in CryptoAlgorithm:
        raise XRPLAddressCodecException(
            f"Encoding type must be one of {CryptoAlgorithm}"
        )

    prefix = _ALGORITHM_TO_PREFIX_MAP[encoding_type][0]
    return _encode(entropy, prefix, SEED_LENGTH)


def decode_seed(
    seed: str, algorithm: Optional[CryptoAlgorithm] = None
) -> Tuple[bytes, CryptoAlgorithm]:
    """
    Returns (decoded seed, its algorithm).

    Args:
        seed: The b58 encoding of a seed.
        algorithm: The encoding algorithm. Inferred from the seed if not included.

    Returns:
        (decoded seed, its algorithm).

    Raises:
        XRPLAddressCodecException: If the seed is invalid.
    """
    if algorithm is not None:
        # check all algorithm prefixes
        for prefix in _ALGORITHM_TO_PREFIX_MAP[algorithm]:
            try:
                decoded_result = _decode(seed, bytes(prefix))
                return decoded_result, algorithm
            except XRPLAddressCodecException:
                # prefix is incorrect, wrong prefix
                continue
        raise XRPLAddressCodecException("Wrong algorithm for the seed type.")

    for algorithm in CryptoAlgorithm:  # use default prefix
        prefix = _ALGORITHM_TO_PREFIX_MAP[algorithm][0]
        try:
            decoded_result = _decode(seed, bytes(prefix))
            return decoded_result, algorithm
        except XRPLAddressCodecException:
            # prefix is incorrect, wrong algorithm
            continue
    raise XRPLAddressCodecException(
        "Invalid seed; could not determine encoding algorithm"
    )


def encode_classic_address(bytestring: bytes) -> str:
    """
    Returns the classic address encoding of these bytes as a base58 string.

    Args:
        bytestring: Bytes to be encoded.

    Returns:
        The classic address encoding of these bytes as a base58 string.
    """
    return _encode(bytestring, _CLASSIC_ADDRESS_PREFIX, _CLASSIC_ADDRESS_LENGTH)


def decode_classic_address(classic_address: str) -> bytes:
    """
    Returns the decoded bytes of the classic address.

    Args:
        classic_address: Classic address to be decoded.

    Returns:
        The decoded bytes of the classic address.
    """
    return _decode(classic_address, bytes(_CLASSIC_ADDRESS_PREFIX))


def encode_node_public_key(bytestring: bytes) -> str:
    """
    Returns the node public key encoding of these bytes as a base58 string.

    Args:
        bytestring: Bytes to be encoded.

    Returns:
        The node public key encoding of these bytes as a base58 string.
    """
    return _encode(bytestring, _NODE_PUBLIC_KEY_PREFIX, _NODE_PUBLIC_KEY_LENGTH)


def decode_node_public_key(node_public_key: str) -> bytes:
    """
    Returns the decoded bytes of the node public key

    Args:
        node_public_key: Node public key to be decoded.

    Returns:
        The decoded bytes of the node public key.

    """
    return _decode(node_public_key, bytes(_NODE_PUBLIC_KEY_PREFIX))


def encode_account_public_key(bytestring: bytes) -> str:
    """
    Returns the account public key encoding of these bytes as a base58 string.

    Args:
        bytestring: Bytes to be encoded.

    Returns:
        The account public key encoding of these bytes as a base58 string.
    """
    return _encode(bytestring, _ACCOUNT_PUBLIC_KEY_PREFIX, _ACCOUNT_PUBLIC_KEY_LENGTH)


def decode_account_public_key(account_public_key: str) -> bytes:
    """
    Returns the decoded bytes of the account public key.

    Args:
        account_public_key: Account public key to be decoded.

    Returns:
        The decoded bytes of the account public key.
    """
    return _decode(account_public_key, bytes(_ACCOUNT_PUBLIC_KEY_PREFIX))


def is_valid_classic_address(classic_address: str) -> bool:
    """
    Returns whether `classic_address` is a valid classic address.

    Args:
        classic_address: The classic address to validate.

    Returns:
        Whether `classic_address` is a valid classic address.
    """
    try:
        decode_classic_address(classic_address)
        return True
    except (XRPLAddressCodecException, ValueError):
        return False



"""Parse offer changes of every offer object involved in the given transaction."""

from typing import List

from xrpl.models import TransactionMetadata
from xrpl.utils.txn_parser.utils import AccountOfferChanges, compute_order_book_changes


def get_order_book_changes(metadata: TransactionMetadata) -> List[AccountOfferChanges]:
    """
    Parse all order book changes from a transaction's metadata.

    Args:
        metadata: Transactions metadata.

    Returns:
        All offer changes caused by the transaction.
        The offer changes are grouped by their owner accounts.
    """
    return compute_order_book_changes(metadata)



"""Model for a XChainCommit transaction type."""

from dataclasses import dataclass, field
from typing import Optional, Union

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainCommit(Transaction):
    """Represents a XChainCommit transaction."""

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore

    xchain_claim_id: Union[int, str] = REQUIRED  # type: ignore

    amount: Amount = REQUIRED  # type: ignore

    other_chain_destination: Optional[str] = None

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_COMMIT,
        init=False,
    )



"""Conversions between XRP drops and native number types."""
import re
from decimal import Decimal, InvalidOperation, localcontext
from typing import Pattern, Union

from typing_extensions import Final

from xrpl.constants import DROPS_DECIMAL_CONTEXT, XRPLException

ONE_DROP: Final[Decimal] = Decimal("0.000001")
"""Indivisible unit of XRP"""

MAX_XRP: Final[Decimal] = Decimal(10**11)
"""100 billion decimal XRP"""

MAX_DROPS: Final[Decimal] = Decimal(10**17)
"""Maximum possible drops of XRP"""

# Drops should be an integer string. MAY have (positive) exponent.
# See also: https://xrpl.org/currency-formats.html#string-numbers
_DROPS_REGEX: Final[Pattern[str]] = re.compile("(?:[1-9][0-9Ee-]{0,17}|0)")


def xrp_to_drops(xrp: Union[int, float, Decimal]) -> str:
    """
    Convert a numeric XRP amount to drops of XRP.

    Args:
        xrp: Numeric representation of whole XRP

    Returns:
        Equivalent amount in drops of XRP

    Raises:
        TypeError: if ``xrp`` is given as a string
        XRPRangeException: if the given amount of XRP is invalid
    """
    if type(xrp) == str:  # type: ignore
        # This protects people from passing drops to this function and getting
        # a million times as many drops back.
        raise TypeError(
            "XRP provided as a string. Use a number format" "like Decimal or int."
        )
    with localcontext(DROPS_DECIMAL_CONTEXT):
        try:
            xrp_d = Decimal(xrp)
        except InvalidOperation:
            raise XRPRangeException(f"Not a valid amount of XRP: '{xrp}'")

        if not xrp_d.is_finite():  # NaN or an Infinity
            raise XRPRangeException(f"Not a valid amount of XRP: '{xrp}'")

        if xrp_d < ONE_DROP and xrp_d != 0:
            raise XRPRangeException(f"XRP amount {xrp} is too small.")
        if xrp_d > MAX_XRP:
            raise XRPRangeException(f"XRP amount {xrp} is too large.")

        drops_amount = (xrp_d / ONE_DROP).quantize(Decimal(1))
        drops_str = str(drops_amount).strip()

        # This should never happen, but is a precaution against Decimal doing
        # something unexpected.
        if not _DROPS_REGEX.fullmatch(drops_str):
            raise XRPRangeException(
                f"xrp_to_drops failed sanity check. Value "
                f"'{drops_str}' does not match the drops regex"
            )

    return drops_str


def drops_to_xrp(drops: str) -> Decimal:
    """
    Convert from drops to decimal XRP.

    Args:
        drops: String representing indivisible drops of XRP

    Returns:
        Decimal representation of the same amount of XRP

    Raises:
        TypeError: if ``drops`` not given as a string
        XRPRangeException: if the given number of drops is invalid
    """
    if type(drops) != str:
        raise TypeError(f"Drops must be provided as string (got {type(drops)})")
    drops = drops.strip()
    with localcontext(DROPS_DECIMAL_CONTEXT):
        if not _DROPS_REGEX.fullmatch(drops):
            raise XRPRangeException(f"Not a valid amount of drops: '{drops}'")
        try:
            drops_d = Decimal(drops)
        except InvalidOperation:
            raise XRPRangeException(f"Not a valid amount of drops: '{drops}'")
        xrp_d = drops_d * ONE_DROP
        if xrp_d > MAX_XRP:
            raise XRPRangeException(f"Drops amount {drops} is too large.")
        return xrp_d


class XRPRangeException(XRPLException):
    """Exception for invalid XRP amounts."""

    pass



"""Model for a XChainCreateBridge transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.currencies import XRP
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainCreateBridge(Transaction):
    """Represents a XChainCreateBridge transaction."""

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore

    signature_reward: str = REQUIRED  # type: ignore

    min_account_create_amount: Optional[str] = None

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_CREATE_BRIDGE,
        init=False,
    )

    def _get_errors(self: XChainCreateBridge) -> Dict[str, str]:
        errors = super()._get_errors()

        bridge = self.xchain_bridge

        if bridge.locking_chain_door == bridge.issuing_chain_door:
            errors[
                "xchain_bridge"
            ] = "Cannot have the same door accounts on the locking and issuing chain."

        if self.account not in [bridge.locking_chain_door, bridge.issuing_chain_door]:
            errors[
                "account"
            ] = "Account must be either locking chain door or issuing chain door."

        if (bridge.locking_chain_issue == XRP()) != (
            bridge.issuing_chain_issue == XRP()
        ):
            errors["issue"] = "Bridge must be XRP-XRP or IOU-IOU."

        if (
            self.min_account_create_amount is not None
            and bridge.locking_chain_issue != XRP()
        ):
            errors[
                "min_account_create_amount"
            ] = "Cannot have MinAccountCreateAmount if bridge is IOU-IOU."

        if self.signature_reward is not None and not self.signature_reward.isnumeric():
            errors["signature_reward"] = "signature_reward must be numeric."

        if (
            self.min_account_create_amount is not None
            and not self.min_account_create_amount.isnumeric()
        ):
            errors[
                "min_account_create_amount_value"
            ] = "min_account_create_amount must be numeric."

        return errors



"""
The ledger_data method retrieves contents of
the specified ledger. You can iterate through
several calls to retrieve the entire contents
of a single ledger version.
`See ledger data <https://xrpl.org/ledger_data.html>`_
"""
from dataclasses import dataclass, field
from typing import Any, Optional, Union

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class LedgerData(Request):
    """
    The ledger_data method retrieves contents of
    the specified ledger. You can iterate through
    several calls to retrieve the entire contents
    of a single ledger version.
    `See ledger data <https://xrpl.org/ledger_data.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER_DATA, init=False)
    ledger_hash: Optional[str] = None
    ledger_index: Optional[Union[str, int]] = None
    binary: bool = False
    limit: Optional[int] = None
    # marker data shape is actually undefined in the spec, up to the
    # implementation of an individual server
    marker: Optional[Any] = None



"""Handles wallet generation from a faucet."""
import asyncio
from typing import Optional

from xrpl.asyncio.wallet import generate_faucet_wallet as async_generate_faucet_wallet
from xrpl.clients.sync_client import SyncClient
from xrpl.wallet.main import Wallet


def generate_faucet_wallet(
    client: SyncClient,
    wallet: Optional[Wallet] = None,
    debug: bool = False,
    faucet_host: Optional[str] = None,
) -> Wallet:
    """
    Generates a random wallet and funds it using the XRPL Testnet Faucet.

    Args:
        client: the network client used to make network calls.
        wallet: the wallet to fund. If omitted or `None`, a new wallet is created.
        debug: Whether to print debug information as it creates the wallet.
        faucet_host: A custom host to use for funding a wallet. In environments other
            than devnet and testnet, this parameter is required.

    Returns:
        A Wallet on the testnet that contains some amount of XRP.

    Raises:
        XRPLFaucetException: if an address could not be funded with the faucet.
        XRPLRequestFailureException: if a request to the ledger fails.
        requests.exceptions.HTTPError: if the request to the faucet fails.

    .. # noqa: DAR402 exception raised in private method
    """
    return asyncio.run(async_generate_faucet_wallet(client, wallet, debug, faucet_host))



"""High-level methods to obtain information about accounts."""

import asyncio
from typing import Dict, Union

from deprecated.sphinx import deprecated

from xrpl.asyncio.account import main
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response


def does_account_exist(
    address: str, client: SyncClient, ledger_index: Union[str, int] = "validated"
) -> bool:
    """
    Query the ledger for whether the account exists.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        Whether the account exists on the ledger.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    return asyncio.run(main.does_account_exist(address, client, ledger_index))


def get_next_valid_seq_number(
    address: str, client: SyncClient, ledger_index: Union[str, int] = "current"
) -> int:
    """
    Query the ledger for the next available sequence number for an account.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "current".

    Returns:
        The next valid sequence number for the address.
    """
    return asyncio.run(main.get_next_valid_seq_number(address, client, ledger_index))


def get_balance(
    address: str, client: SyncClient, ledger_index: Union[str, int] = "validated"
) -> int:
    """
    Query the ledger for the balance of the given account.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        The balance of the address.
    """
    return asyncio.run(main.get_balance(address, client, ledger_index))


def get_account_root(
    address: str, client: SyncClient, ledger_index: Union[str, int] = "validated"
) -> Dict[str, Union[int, str]]:
    """
    Query the ledger for the AccountRoot object associated with a given address.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        The AccountRoot dictionary for the address.
    """
    return asyncio.run(main.get_account_root(address, client, ledger_index))


@deprecated(
    reason="Sending an AccountInfo request directly is just as easy to use.",
    version="1.8.0",
)
def get_account_info(
    address: str, client: SyncClient, ledger_index: Union[str, int] = "validated"
) -> Response:
    """
    Query the ledger for account info of given address.

    Args:
        address: the account to query.
        client: the network client used to make network calls.
        ledger_index: The ledger index to use for the request. Must be an integer
            ledger value or "current" (the current working version), "closed" (for the
            closed-and-proposed version), or "validated" (the most recent version
            validated by consensus). The default is "validated".

    Returns:
        The account info for the address.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    return asyncio.run(main.get_account_info(address, client, ledger_index))



"""Models for a transaction's metadata."""

from typing import Dict, List, Union

from typing_extensions import Literal, NotRequired, TypedDict

from xrpl.models.amounts.amount import Amount


class Fields(TypedDict):
    """Model for possible fields."""

    Flags: int
    Sequence: int
    Account: NotRequired[str]
    LowLimit: NotRequired[Dict[str, str]]
    HighLimit: NotRequired[Dict[str, str]]
    Balance: NotRequired[Union[Dict[str, str], str]]
    TakerGets: NotRequired[Union[Dict[str, str], str]]
    TakerPays: NotRequired[Union[Dict[str, str], str]]
    BookDirectory: NotRequired[str]
    Expiration: NotRequired[int]


class CreatedNodeFields(TypedDict):
    """Fields of a CreatedNode."""

    LedgerEntryType: str
    LedgerIndex: str
    NewFields: Fields


class CreatedNode(TypedDict):
    """CreatedNode model."""

    CreatedNode: CreatedNodeFields


class ModifiedNodeFields(TypedDict):
    """Fields of a ModifiedNode."""

    LedgerEntryType: str
    LedgerIndex: str
    FinalFields: NotRequired[Fields]
    PreviousFields: NotRequired[Fields]
    PreviousTxnID: NotRequired[str]
    PreviousTxnLgrSeq: NotRequired[int]


class ModifiedNode(TypedDict):
    """ModifiedNode model."""

    ModifiedNode: ModifiedNodeFields


class DeletedNodeFields(TypedDict):
    """Fields of a DeletedNode."""

    LedgerEntryType: str
    LedgerIndex: str
    FinalFields: Fields
    PreviousFields: NotRequired[Fields]


class DeletedNode(TypedDict):
    """DeletedNode model."""

    DeletedNode: DeletedNodeFields


class TransactionMetadata(TypedDict):
    """A model for a transaction's metadata."""

    AffectedNodes: List[Union[CreatedNode, ModifiedNode, DeletedNode]]
    TransactionIndex: int
    TransactionResult: str
    DeliveredAmount: NotRequired[Amount]
    # "unavailable" possible for transactions before 2014-01-20
    delivered_amount: NotRequired[Union[Amount, Literal["unavailable"]]]



"""Base class for XRPL Hash types.
`See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
"""
from __future__ import annotations  # Requires Python 3.7+

from abc import ABC, abstractmethod
from typing import Optional, Type

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.serialized_type import SerializedType


class Hash(SerializedType, ABC):
    """
    Base class for XRPL Hash types.
    `See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
    """

    def __init__(self: Hash, buffer: Optional[bytes]) -> None:
        """
        Construct a Hash.

        Args:
            buffer: The byte buffer that will be used to store the serialized
                encoding of this field.
        """
        buffer = buffer if buffer is not None else bytes(self._get_length())

        if len(buffer) != self._get_length():
            raise XRPLBinaryCodecException(
                f"Invalid hash length {len(buffer)}. Expected {self._get_length()}"
            )
        super().__init__(buffer)

    def __str__(self: Hash) -> str:
        """Returns a hex-encoded string representation of the bytes buffer."""
        return self.to_hex()

    @classmethod
    def from_value(cls: Type[Hash], value: str) -> Hash:
        """
        Construct a Hash object from a hex string.

        Args:
            value: The value to construct a Hash from.

        Returns:
            The Hash object constructed from value.

        Raises:
            XRPLBinaryCodecException: If the supplied value is of the wrong type.
        """
        if not isinstance(value, str):
            raise XRPLBinaryCodecException(
                f"Invalid type to construct a {cls.__name__}: expected str,"
                f" received {value.__class__.__name__}."
            )

        return cls(bytes.fromhex(value))

    @classmethod
    def from_parser(
        cls: Type[Hash], parser: BinaryParser, length_hint: Optional[int] = None
    ) -> Hash:
        """
        Construct a Hash object from an existing BinaryParser.

        Args:
            parser: The parser to construct the Hash object from.
            length_hint: The number of bytes to consume from the parser.

        Returns:
            The Hash object constructed from a parser.
        """
        num_bytes = length_hint if length_hint is not None else cls._get_length()
        return cls(parser.read(num_bytes))

    @classmethod
    @abstractmethod
    def _get_length(cls: Type[Hash]) -> int:
        pass



"""Model for SetFee pseudo-transaction type."""

from dataclasses import dataclass, field
from typing import Optional

from xrpl.models.required import REQUIRED
from xrpl.models.transactions.pseudo_transactions.pseudo_transaction import (
    PseudoTransaction,
)
from xrpl.models.transactions.types import PseudoTransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class SetFee(PseudoTransaction):
    """
    A SetFee pseudo-transaction marks a change in `transaction cost
    <https://xrpl.org/transaction-cost.html>`_ or `reserve requirements
    <https://xrpl.org/reserves.html>`_ as a result of `Fee Voting
    <https://xrpl.org/fee-voting.html>`_.
    """

    base_fee: str = REQUIRED  # type: ignore
    """
    The charge, in drops of XRP, for the reference transaction, as hex. (This is the
    transaction cost before scaling for load.) This field is required.

    :meta hide-value:
    """

    reference_fee_units: int = REQUIRED  # type: ignore
    """
    The cost, in fee units, of the reference transaction. This field is required.

    :meta hide-value:
    """

    reserve_base: int = REQUIRED  # type: ignore
    """
    The base reserve, in drops. This field is required.

    :meta hide-value:
    """

    reserve_increment: int = REQUIRED  # type: ignore
    """
    The incremental reserve, in drops. This field is required.

    :meta hide-value:
    """

    ledger_sequence: Optional[int] = None
    """
    The index of the ledger version where this pseudo-transaction appears. This
    distinguishes the pseudo-transaction from other occurrences of the same change.
    This field is omitted for some historical SetFee pseudo-transactions.
    """

    transaction_type: PseudoTransactionType = field(
        default=PseudoTransactionType.SET_FEE,
        init=False,
    )



"""Enum containing the different Transaction types."""

from enum import Enum


class TransactionType(str, Enum):
    """Enum containing the different Transaction types."""

    ACCOUNT_DELETE = "AccountDelete"
    ACCOUNT_SET = "AccountSet"
    AMM_BID = "AMMBid"
    AMM_CREATE = "AMMCreate"
    AMM_DEPOSIT = "AMMDeposit"
    AMM_VOTE = "AMMVote"
    AMM_WITHDRAW = "AMMWithdraw"
    CHECK_CANCEL = "CheckCancel"
    CHECK_CASH = "CheckCash"
    CHECK_CREATE = "CheckCreate"
    DEPOSIT_PREAUTH = "DepositPreauth"
    ESCROW_CANCEL = "EscrowCancel"
    ESCROW_CREATE = "EscrowCreate"
    ESCROW_FINISH = "EscrowFinish"
    NFTOKEN_ACCEPT_OFFER = "NFTokenAcceptOffer"
    NFTOKEN_BURN = "NFTokenBurn"
    NFTOKEN_CANCEL_OFFER = "NFTokenCancelOffer"
    NFTOKEN_CREATE_OFFER = "NFTokenCreateOffer"
    NFTOKEN_MINT = "NFTokenMint"
    OFFER_CANCEL = "OfferCancel"
    OFFER_CREATE = "OfferCreate"
    PAYMENT = "Payment"
    PAYMENT_CHANNEL_CLAIM = "PaymentChannelClaim"
    PAYMENT_CHANNEL_CREATE = "PaymentChannelCreate"
    PAYMENT_CHANNEL_FUND = "PaymentChannelFund"
    SET_REGULAR_KEY = "SetRegularKey"
    SIGNER_LIST_SET = "SignerListSet"
    TICKET_CREATE = "TicketCreate"
    TRUST_SET = "TrustSet"
    XCHAIN_ACCOUNT_CREATE_COMMIT = "XChainAccountCreateCommit"
    XCHAIN_ADD_ACCOUNT_CREATE_ATTESTATION = "XChainAddAccountCreateAttestation"
    XCHAIN_ADD_CLAIM_ATTESTATION = "XChainAddClaimAttestation"
    XCHAIN_CLAIM = "XChainClaim"
    XCHAIN_COMMIT = "XChainCommit"
    XCHAIN_CREATE_BRIDGE = "XChainCreateBridge"
    XCHAIN_CREATE_CLAIM_ID = "XChainCreateClaimID"
    XCHAIN_MODIFY_BRIDGE = "XChainModifyBridge"



"""
WebSocket API only! The path_find method searches for a
path along which a transaction can possibly be made, and
periodically sends updates when the path changes over time.
For a simpler version that is supported by JSON-RPC, see the
ripple_path_find method. For payments occurring strictly in XRP,
it is not necessary to find a path, because XRP can be sent
directly to any account.

There are three different modes, or sub-commands, of the path_find
command. Specify which one you want with the subcommand parameter:

create - Start sending pathfinding information
close - Stop sending pathfinding information
status - Get the information of the currently-open pathfinding request
Although the rippled server tries to find the cheapest path or combination
of paths for making a payment, it is not guaranteed that the paths returned
by this method are, in fact, the best paths. Due to server load,
pathfinding may not find the best results. Additionally, you should be
careful with the pathfinding results from untrusted servers. A server
could be modified to return less-than-optimal paths to earn money for its
operators. If you do not have your own server that you can trust with
pathfinding, you should compare the results of pathfinding from multiple
servers run by different parties, to minimize the risk of a single server
returning poor results. (Note: A server returning less-than-optimal
results is not necessarily proof of malicious behavior; it could also be
a symptom of heavy server load.)
"""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

from xrpl.models.amounts import Amount
from xrpl.models.path import Path
from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


class PathFindSubcommand(str, Enum):
    """
    There are three different modes, or sub-commands, of the path_find
    command. Specify which one you want with the subcommand parameter:

    create - Start sending pathfinding information
    close - Stop sending pathfinding information
    status - Get the information of the currently-open pathfinding request
    """

    CREATE = "create"
    CLOSE = "close"
    STATUS = "status"


@require_kwargs_on_init
@dataclass(frozen=True)
class PathFind(Request):
    """
    WebSocket API only! The path_find method searches for a
    path along which a transaction can possibly be made, and
    periodically sends updates when the path changes over time.
    For a simpler version that is supported by JSON-RPC, see the
    ripple_path_find method. For payments occurring strictly in XRP,
    it is not necessary to find a path, because XRP can be sent
    directly to any account.

    Although the rippled server tries to find the cheapest path or combination
    of paths for making a payment, it is not guaranteed that the paths returned
    by this method are, in fact, the best paths. Due to server load,
    pathfinding may not find the best results. Additionally, you should be
    careful with the pathfinding results from untrusted servers. A server
    could be modified to return less-than-optimal paths to earn money for its
    operators. If you do not have your own server that you can trust with
    pathfinding, you should compare the results of pathfinding from multiple
    servers run by different parties, to minimize the risk of a single server
    returning poor results. (Note: A server returning less-than-optimal
    results is not necessarily proof of malicious behavior; it could also be
    a symptom of heavy server load.)
    """

    subcommand: PathFindSubcommand = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    source_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    destination_amount: Amount = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.PATH_FIND, init=False)
    send_max: Optional[Amount] = None
    paths: Optional[List[Path]] = None



"""
Model objects for specific `types of Transactions
<https://xrpl.org/transaction-types.html>`_ in the XRP Ledger.
"""

from xrpl.models.transactions.account_delete import AccountDelete
from xrpl.models.transactions.account_set import (
    AccountSet,
    AccountSetFlag,
    AccountSetFlagInterface,
)
from xrpl.models.transactions.amm_bid import AMMBid, AuthAccount
from xrpl.models.transactions.amm_create import AMMCreate
from xrpl.models.transactions.amm_deposit import (
    AMMDeposit,
    AMMDepositFlag,
    AMMDepositFlagInterface,
)
from xrpl.models.transactions.amm_vote import AMMVote
from xrpl.models.transactions.amm_withdraw import (
    AMMWithdraw,
    AMMWithdrawFlag,
    AMMWithdrawFlagInterface,
)
from xrpl.models.transactions.check_cancel import CheckCancel
from xrpl.models.transactions.check_cash import CheckCash
from xrpl.models.transactions.check_create import CheckCreate
from xrpl.models.transactions.deposit_preauth import DepositPreauth
from xrpl.models.transactions.escrow_cancel import EscrowCancel
from xrpl.models.transactions.escrow_create import EscrowCreate
from xrpl.models.transactions.escrow_finish import EscrowFinish
from xrpl.models.transactions.metadata import TransactionMetadata
from xrpl.models.transactions.nftoken_accept_offer import NFTokenAcceptOffer
from xrpl.models.transactions.nftoken_burn import NFTokenBurn
from xrpl.models.transactions.nftoken_cancel_offer import NFTokenCancelOffer
from xrpl.models.transactions.nftoken_create_offer import (
    NFTokenCreateOffer,
    NFTokenCreateOfferFlag,
    NFTokenCreateOfferFlagInterface,
)
from xrpl.models.transactions.nftoken_mint import (
    NFTokenMint,
    NFTokenMintFlag,
    NFTokenMintFlagInterface,
)
from xrpl.models.transactions.offer_cancel import OfferCancel
from xrpl.models.transactions.offer_create import (
    OfferCreate,
    OfferCreateFlag,
    OfferCreateFlagInterface,
)
from xrpl.models.transactions.payment import Payment, PaymentFlag, PaymentFlagInterface
from xrpl.models.transactions.payment_channel_claim import (
    PaymentChannelClaim,
    PaymentChannelClaimFlag,
    PaymentChannelClaimFlagInterface,
)
from xrpl.models.transactions.payment_channel_create import PaymentChannelCreate
from xrpl.models.transactions.payment_channel_fund import PaymentChannelFund
from xrpl.models.transactions.set_regular_key import SetRegularKey
from xrpl.models.transactions.signer_list_set import SignerEntry, SignerListSet
from xrpl.models.transactions.ticket_create import TicketCreate
from xrpl.models.transactions.transaction import Memo, Signer, Transaction
from xrpl.models.transactions.trust_set import (
    TrustSet,
    TrustSetFlag,
    TrustSetFlagInterface,
)
from xrpl.models.transactions.xchain_account_create_commit import (
    XChainAccountCreateCommit,
)
from xrpl.models.transactions.xchain_add_account_create_attestation import (
    XChainAddAccountCreateAttestation,
)
from xrpl.models.transactions.xchain_add_claim_attestation import (
    XChainAddClaimAttestation,
)
from xrpl.models.transactions.xchain_claim import XChainClaim
from xrpl.models.transactions.xchain_commit import XChainCommit
from xrpl.models.transactions.xchain_create_bridge import XChainCreateBridge
from xrpl.models.transactions.xchain_create_claim_id import XChainCreateClaimID
from xrpl.models.transactions.xchain_modify_bridge import (
    XChainModifyBridge,
    XChainModifyBridgeFlag,
    XChainModifyBridgeFlagInterface,
)

__all__ = [
    "AccountDelete",
    "AccountSet",
    "AccountSetFlag",
    "AccountSetFlagInterface",
    "AMMBid",
    "AMMCreate",
    "AMMDeposit",
    "AMMDepositFlag",
    "AMMDepositFlagInterface",
    "AMMVote",
    "AMMWithdraw",
    "AMMWithdrawFlag",
    "AMMWithdrawFlagInterface",
    "AuthAccount",
    "CheckCancel",
    "CheckCash",
    "CheckCreate",
    "DepositPreauth",
    "EscrowCancel",
    "EscrowCreate",
    "EscrowFinish",
    "Memo",
    "NFTokenAcceptOffer",
    "NFTokenBurn",
    "NFTokenCancelOffer",
    "NFTokenCreateOffer",
    "NFTokenCreateOfferFlag",
    "NFTokenCreateOfferFlagInterface",
    "NFTokenMint",
    "NFTokenMintFlag",
    "NFTokenMintFlagInterface",
    "OfferCancel",
    "OfferCreate",
    "OfferCreateFlag",
    "OfferCreateFlagInterface",
    "Payment",
    "PaymentFlag",
    "PaymentFlagInterface",
    "PaymentChannelClaim",
    "PaymentChannelClaimFlag",
    "PaymentChannelClaimFlagInterface",
    "PaymentChannelCreate",
    "PaymentChannelFund",
    "SetRegularKey",
    "Signer",
    "SignerEntry",
    "SignerListSet",
    "TicketCreate",
    "Transaction",
    "TransactionMetadata",
    "TrustSet",
    "TrustSetFlag",
    "TrustSetFlagInterface",
    "XChainAccountCreateCommit",
    "XChainAddAccountCreateAttestation",
    "XChainAddClaimAttestation",
    "XChainClaim",
    "XChainCommit",
    "XChainCreateBridge",
    "XChainCreateClaimID",
    "XChainModifyBridge",
    "XChainModifyBridgeFlag",
    "XChainModifyBridgeFlagInterface",
]



"""Model for AMMWithdraw transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from xrpl.models.amounts import Amount, IssuedCurrencyAmount
from xrpl.models.currencies import Currency
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class AMMWithdrawFlag(int, Enum):
    """
    Transactions of the AMMWithdraw type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_LP_TOKEN = 0x00010000
    TF_WITHDRAW_ALL = 0x00020000
    TF_ONE_ASSET_WITHDRAW_ALL = 0x00040000
    TF_SINGLE_ASSET = 0x00080000
    TF_TWO_ASSET = 0x00100000
    TF_ONE_ASSET_LP_TOKEN = 0x00200000
    TF_LIMIT_LP_TOKEN = 0x00400000


class AMMWithdrawFlagInterface(FlagInterface):
    """
    Transactions of the AMMWithdraw type support additional values in the Flags field.
    This TypedDict represents those options.
    """

    TF_LP_TOKEN: bool
    TF_WITHDRAW_ALL: bool
    TF_ONE_ASSET_WITHDRAW_ALL: bool
    TF_SINGLE_ASSET: bool
    TF_TWO_ASSET: bool
    TF_ONE_ASSET_LP_TOKEN: bool
    TF_LIMIT_LP_TOKEN: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class AMMWithdraw(Transaction):
    """
    AMMWithdraw is the withdraw transaction used to remove liquidity from the AMM
    instance pool, thus redeeming some share of the pools that one owns in the form
    of LPToken.

    The following are the recommended valid combinations:
    - LPTokenIn
    - Amount
    - Amount and Amount2
    - Amount and LPTokenIn
    - Amount and EPrice
    """

    asset: Currency = REQUIRED  # type: ignore
    """
    Specifies one of the pool assets (XRP or token) of the AMM instance.
    """

    asset2: Currency = REQUIRED  # type: ignore
    """
    Specifies the other pool asset of the AMM instance.
    """

    lp_token_in: Optional[IssuedCurrencyAmount] = None
    """
    Specifies the amount of shares of the AMM instance pools that the trader
    wants to redeem or trade in.
    """

    amount: Optional[Amount] = None
    """
    Specifies one of the pools assets that the trader wants to remove.
    If the asset is XRP, then the Amount is a string specifying the number of drops.
    Otherwise it is an IssuedCurrencyAmount object.
    """

    amount2: Optional[Amount] = None
    """
    Specifies the other pool asset that the trader wants to remove.
    """

    e_price: Optional[Amount] = None
    """
    Specifies the effective-price of the token out after successful execution of
    the transaction.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.AMM_WITHDRAW,
        init=False,
    )

    def _get_errors(self: AMMWithdraw) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.amount2 is not None and self.amount is None:
            errors["AMMWithdraw"] = "Must set `amount` with `amount2`"
        elif self.e_price is not None and self.amount is None:
            errors["AMMWithdraw"] = "Must set `amount` with `e_price`"
        return errors



"""
The transaction_entry method retrieves information on a single transaction from a
specific ledger version. (The tx method, by contrast, searches all ledgers for the
specified transaction. We recommend using that method instead.)

`See transaction_entry <https://xrpl.org/transaction_entry.html>`_
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Union

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class TransactionEntry(Request):
    """
    The transaction_entry method retrieves information on a single transaction from a
    specific ledger version. (The tx method, by contrast, searches all ledgers for the
    specified transaction. We recommend using that method instead.)

    `See transaction_entry <https://xrpl.org/transaction_entry.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.TRANSACTION_ENTRY, init=False)
    ledger_hash: Optional[str] = None
    ledger_index: Optional[Union[str, int]] = None
    tx_hash: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """



"""This module handles everything related to X-Addresses."""

from typing import Optional, Tuple

import base58
from typing_extensions import Final

from xrpl.core.addresscodec.codec import decode_classic_address, encode_classic_address
from xrpl.core.addresscodec.exceptions import XRPLAddressCodecException
from xrpl.core.addresscodec.utils import XRPL_ALPHABET

MAX_32_BIT_UNSIGNED_INT: Final[int] = 4294967295

_PREFIX_BYTES_MAIN: Final[bytes] = bytes([0x05, 0x44])  # 5, 68
_PREFIX_BYTES_TEST: Final[bytes] = bytes([0x04, 0x93])  # 4, 147

# To better understand the cryptographic details, visit
# https://github.com/xrp-community/standards-drafts/issues/6

# General format of an X-Address:
# [← 2 byte prefix →|← 160 bits of account ID →|← 8 bits of flags →|← 64 bits of tag →]


def classic_address_to_xaddress(
    classic_address: str, tag: Optional[int], is_test_network: bool
) -> str:
    """
    Returns the X-Address representation of the data.

    Args:
        classic_address: The base58 encoding of the classic address.
        tag: The destination tag.
        is_test_network: Whether it is the test network or the main network.

    Returns:
        The X-Address representation of the data.

    Raises:
        XRPLAddressCodecException: If the classic address does not have enough bytes
            or the tag is invalid.
    """
    classic_address_bytes = decode_classic_address(classic_address)
    if len(classic_address_bytes) != 20:
        raise XRPLAddressCodecException("Account ID must be 20 bytes")

    if tag is not None and tag > MAX_32_BIT_UNSIGNED_INT:
        raise XRPLAddressCodecException("Invalid tag")

    flag = tag is not None
    if tag is None:
        tag = 0

    bytestring = _PREFIX_BYTES_TEST if is_test_network else _PREFIX_BYTES_MAIN
    bytestring += classic_address_bytes
    encoded_tag = bytes(
        [
            flag,
            tag & 0xFF,
            tag >> 8 & 0xFF,
            tag >> 16 & 0xFF,
            tag >> 24 & 0xFF,
            0,
            0,
            0,
            0,
        ]
    )
    bytestring += encoded_tag

    return base58.b58encode_check(bytestring, alphabet=XRPL_ALPHABET).decode("utf-8")


def xaddress_to_classic_address(xaddress: str) -> Tuple[str, Optional[int], bool]:
    """
    Returns a tuple containing the classic address, tag, and whether the address
    is on a test network for an X-Address.

    Args:
        xaddress: base58-encoded X-Address.

    Returns:
        A tuple containing:
            classic_address: the base58 classic address
            tag: the destination tag
            is_test_network: whether the address is on the test network (or main)
    """
    decoded = base58.b58decode_check(
        xaddress, alphabet=XRPL_ALPHABET
    )  # convert b58 to bytes
    is_test_network = _is_test_address(decoded[:2])
    classic_address_bytes = decoded[2:22]
    tag = _get_tag_from_buffer(decoded[22:])  # extracts the destination tag

    classic_address = encode_classic_address(classic_address_bytes)
    return (classic_address, tag, is_test_network)


def _is_test_address(prefix: bytes) -> bool:
    """
    Returns whether a decoded X-Address is a test address.

    Args:
        prefix: The first 2 bytes of an X-Address.

    Returns:
        Whether a decoded X-Address is a test address.

    Raises:
        XRPLAddressCodecException: If the prefix is invalid.
    """
    if _PREFIX_BYTES_MAIN == prefix:
        return False
    if _PREFIX_BYTES_TEST == prefix:
        return True
    raise XRPLAddressCodecException("Invalid X-Address: bad prefix")


def _get_tag_from_buffer(buffer: bytes) -> Optional[int]:
    """
    Returns the destination tag extracted from the suffix of the X-Address.

    Args:
        buffer: The buffer to extract a destination tag from.

    Returns:
        The destination tag extracted from the suffix of the X-Address.
    """
    flag = buffer[0]
    if flag >= 2:
        raise XRPLAddressCodecException("Unsupported X-Address")
    if flag == 1:  # Little-endian to big-endian
        return (
            buffer[1] + buffer[2] * 0x100 + buffer[3] * 0x10000 + buffer[4] * 0x1000000
        )  # inverse of what happens in encode
    if flag != 0:
        raise XRPLAddressCodecException("Flag must be zero to indicate no tag")
    if bytes.fromhex("0000000000000000") != buffer[1:9]:
        raise XRPLAddressCodecException("Remaining bytes must be zero")
    return None


def is_valid_xaddress(xaddress: str) -> bool:
    """
    Returns whether ``xaddress`` is a valid X-Address.

    Args:
        xaddress: The X-Address to check for validity.

    Returns:
        Whether ``xaddress`` is a valid X-Address.
    """
    try:
        xaddress_to_classic_address(xaddress)
        return True
    except (XRPLAddressCodecException, ValueError):
        return False



"""High-level ledger methods with the XRPL ledger."""

import asyncio
from typing import Optional

from xrpl.asyncio.ledger import main
from xrpl.clients.sync_client import SyncClient


def get_latest_validated_ledger_sequence(client: SyncClient) -> int:
    """
    Returns the sequence number of the latest validated ledger.

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest validated ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    return asyncio.run(main.get_latest_validated_ledger_sequence(client))


def get_latest_open_ledger_sequence(client: SyncClient) -> int:
    """
    Returns the sequence number of the latest open ledger.

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest open ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    return asyncio.run(main.get_latest_open_ledger_sequence(client))


def get_fee(
    client: SyncClient,
    *,
    max_fee: Optional[float] = 2,
    fee_type: str = "open",
) -> str:
    """
    Query the ledger for the current transaction fee.

    Args:
        client: the network client used to make network calls.
        max_fee: The maximum fee in XRP that the user wants to pay. If load gets too
            high, then the fees will not scale past the maximum fee. If None, there is
            no ceiling for the fee. The default is 2 XRP.
        fee_type: The type of fee to return. The options are "open" (the load-scaled
            fee to get into the open ledger), "minimum" (the minimum transaction
            fee) or "dynamic" (dynamic fee-calculation based on the queue size
            of the node). The default is "open". The recommended option is
            "dynamic".

    Returns:
        The transaction fee, in drops.
        `Read more about drops <https://xrpl.org/currency-formats.html#xrp-amounts>`_

    Raises:
        XRPLException: if an incorrect option for `fee_type` is passed in.
        XRPLRequestFailureException: if the rippled API call fails.
    """
    return asyncio.run(main.get_fee(client, max_fee=max_fee, fee_type=fee_type))



"""
This request returns information about an account's trust lines, including balances
in all non-XRP currencies and assets. All information retrieved is relative to a
particular version of the ledger.

`See account_lines <https://xrpl.org/account_lines.html>`_
"""
from dataclasses import dataclass, field
from typing import Any, Optional, Union

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AccountLines(Request):
    """
    This request returns information about an account's trust lines, including balances
    in all non-XRP currencies and assets. All information retrieved is relative to a
    particular version of the ledger.

    `See account_lines <https://xrpl.org/account_lines.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    ledger_hash: Optional[str] = None
    ledger_index: Optional[Union[str, int]] = None
    method: RequestMethod = field(default=RequestMethod.ACCOUNT_LINES, init=False)
    peer: Optional[str] = None
    limit: Optional[int] = None
    # marker data shape is actually undefined in the spec, up to the
    # implementation of an individual server
    marker: Optional[Any] = None



"""
Represents a TrustSet transaction on the XRP Ledger.
Creates or modifies a trust line linking two accounts.

`See TrustSet <https://xrpl.org/trustset.html>`_
"""
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from xrpl.models.amounts import IssuedCurrencyAmount
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class TrustSetFlag(int, Enum):
    """
    Transactions of the TrustSet type support additional values in the Flags field.
    This enum represents those options.
    """

    TF_SET_AUTH = 0x00010000
    """
    Authorize the other party to hold
    `currency issued by this account <https://xrpl.org/tokens.html>`_.
    (No effect unless using the `asfRequireAuth AccountSet flag
    <https://xrpl.org/accountset.html#accountset-flags>`_.) Cannot be unset.
    """

    TF_SET_NO_RIPPLE = 0x00020000
    """
    Enable the No Ripple flag, which blocks
    `rippling <https://xrpl.org/rippling.html>`_ between two trust
    lines of the same currency if this flag is enabled on both.
    """

    TF_CLEAR_NO_RIPPLE = 0x00040000
    """Disable the No Ripple flag, allowing rippling on this trust line."""

    TF_SET_FREEZE = 0x00100000
    """Freeze the trust line."""

    TF_CLEAR_FREEZE = 0x00200000
    """Unfreeze the trust line."""


class TrustSetFlagInterface(FlagInterface):
    """
    Transactions of the TrustSet type support additional values in the Flags field.
    This TypedDict represents those options.
    """

    TF_SET_AUTH: bool
    TF_SET_NO_RIPPLE: bool
    TF_CLEAR_NO_RIPPLE: bool
    TF_SET_FREEZE: bool
    TF_CLEAR_FREEZE: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class TrustSet(Transaction):
    """
    Represents a TrustSet transaction on the XRP Ledger.
    Creates or modifies a trust line linking two accounts.

    `See TrustSet <https://xrpl.org/trustset.html>`_
    """

    limit_amount: IssuedCurrencyAmount = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    quality_in: Optional[int] = None

    quality_out: Optional[int] = None

    transaction_type: TransactionType = field(
        default=TransactionType.TRUST_SET,
        init=False,
    )



"""
Specifies XRP as a currency, without a value. Normally, you will not use this
model as it does not specify an amount of XRP. In cases where you need to
specify an amount of XRP, you will use a string. However, for some book order
requests where currencies are specified without amounts, you may need to
specify the use of XRP, without a value. In these cases, you will use this
object.

See https://xrpl.org/currency-formats.html#specifying-currency-amounts
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Type, Union

from xrpl.models.base_model import BaseModel
from xrpl.models.exceptions import XRPLModelException
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class XRP(BaseModel):
    """
    Specifies XRP as a currency, without a value. Normally, you will not use this
    model as it does not specify an amount of XRP. In cases where you need to
    specify an amount of XRP, you will use a string. However, for some book order
    requests where currencies are specified without amounts, you may need to
    specify the use of XRP, without a value. In these cases, you will use this
    object.

    See https://xrpl.org/currency-formats.html#specifying-currency-amounts
    """

    currency: str = field(default="XRP", init=False)

    @classmethod
    def from_dict(cls: Type[XRP], value: Dict[str, Any]) -> XRP:
        """
        Construct a new XRP from a dictionary of parameters.

        Args:
            value: The value to construct the XRP from.

        Returns:
            A new XRP object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        if len(value) != 1 or "currency" not in value or value["currency"] != "XRP":
            raise XRPLModelException("Not a valid XRP type")
        return XRP()

    def to_dict(self: XRP) -> Dict[str, Any]:
        """
        Returns the dictionary representation of an XRP currency object.

        Returns:
            The dictionary representation of an XRP currency object.
        """
        return {**super().to_dict(), "currency": "XRP"}

    def to_amount(self: XRP, value: Union[str, int]) -> str:
        """
        Converts value to XRP.

        Args:
            value: The amount of XRP.

        Returns:
            A string representation of XRP amount.
        """
        return str(value)



"""Methods for working with transactions on the XRP Ledger."""
from xrpl.asyncio.transaction import (
    XRPLReliableSubmissionException,
    transaction_json_to_binary_codec_form,
)
from xrpl.transaction.ledger import get_transaction_from_hash
from xrpl.transaction.main import (
    autofill,
    autofill_and_sign,
    safe_sign_and_autofill_transaction,
    safe_sign_and_submit_transaction,
    safe_sign_transaction,
    sign,
    sign_and_submit,
    submit,
    submit_transaction,
)
from xrpl.transaction.multisign import multisign
from xrpl.transaction.reliable_submission import send_reliable_submission

__all__ = [
    "autofill",
    "autofill_and_sign",
    "get_transaction_from_hash",
    "safe_sign_transaction",
    "safe_sign_and_autofill_transaction",
    "safe_sign_and_submit_transaction",
    "sign",
    "sign_and_submit",
    "submit",
    "submit_transaction",
    "transaction_json_to_binary_codec_form",
    "send_reliable_submission",
    "multisign",
    "XRPLReliableSubmissionException",
]



"""
This request retrieves information about an account, its activity, and its XRP
balance.

All information retrieved is relative to a particular version of the ledger.

`See account_info <https://xrpl.org/account_info.html>`_
"""
from dataclasses import dataclass, field
from typing import Optional, Union

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AccountInfo(Request):
    """
    This request retrieves information about an account, its activity, and its XRP
    balance.

    All information retrieved is relative to a particular version of the ledger.

    `See account_info <https://xrpl.org/account_info.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    ledger_hash: Optional[str] = None
    ledger_index: Optional[Union[str, int]] = None
    method: RequestMethod = field(default=RequestMethod.ACCOUNT_INFO, init=False)
    queue: bool = False
    signer_lists: bool = False
    strict: bool = False



"""Model for PaymentChannelClaim transaction type."""
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from xrpl.models.amounts import Amount
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


class PaymentChannelClaimFlag(int, Enum):
    """
    Transactions of the PaymentChannelClaim type support additional values in the Flags
    field. This enum represents those options.

    `See PaymentChannelClaim Flags
    <https://xrpl.org/paymentchannelclaim.html#paymentchannelclaim-flags>`_
    """

    TF_RENEW = 0x00010000
    """
    Clear the channel's `Expiration` time. (`Expiration` is different from the
    channel's immutable `CancelAfter` time.) Only the source address of the payment
    channel can use this flag.
    """

    TF_CLOSE = 0x00020000
    """
    Request to close the channel. Only the channel source and destination addresses
    can use this flag. This flag closes the channel immediately if it has no more
    funds allocated to it after processing the current claim, or if the destination
    address uses it. If the source address uses this flag when the channel still
    holds a value, this schedules the channel to close after `SettleDelay` seconds have
    passed. (Specifically, this sets the `Expiration` of the channel to the close
    time of the previous ledger plus the channel's `SettleDelay` time, unless the
    channel already has an earlier `Expiration` time.) If the destination address
    uses this flag when the channel still holds an amount, any amount that remains
    after processing the claim is returned to the source address.
    """


class PaymentChannelClaimFlagInterface(FlagInterface):
    """
    Transactions of the PaymentChannelClaim type support additional values in the Flags
    field. This TypedDict represents those options.

    `See PaymentChannelClaim Flags
    <https://xrpl.org/paymentchannelclaim.html#paymentchannelclaim-flags>`_
    """

    TF_RENEW: bool
    TF_CLOSE: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class PaymentChannelClaim(Transaction):
    """
    Represents a `PaymentChannelClaim <https://xrpl.org/paymentchannelclaim.html>`_
    transaction, which claims an amount from a `payment channel
    <https://xrpl.org/payment-channels.html>`_, adjusts
    channel's expiration, or both. This transaction can be used differently
    depending on the transaction sender's role in the specified channel.
    """

    channel: str = REQUIRED  # type: ignore
    """
    The unique ID of the payment channel, as a 64-character hexadecimal
    string. This field is required.

    :meta hide-value:
    """

    balance: Optional[Amount] = None
    """
    Total amount delivered by this channel after processing this claim. Required
    to deliver amount. Must be more than the total amount delivered by the channel
    so far, but not greater than the Amount of the signed claim. Must be provided
    except when closing the channel.
    """

    amount: Optional[Amount] = None
    """
    The amount authorized by the Signature. This must match the amount in the signed
    message. This is the cumulative amount that can be dispensed by the channel,
    including amounts previously redeemed. Required unless closing the channel.
    """

    signature: Optional[str] = None
    """
    The signature of the claim, as hexadecimal. This signature must be
    verifiable for the this channel and the given ``public_key`` and ``amount``
    values. May be omitted if closing the channel or if the sender of this
    transaction is the source address of the channel; required otherwise.
    """

    public_key: Optional[str] = None
    """
    The public key that should be used to verify the attached signature. Must
    match the `PublicKey` that was provided when the channel was created.
    Required if ``signature`` is provided.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.PAYMENT_CHANNEL_CLAIM,
        init=False,
    )



"""
The `nft_sell_offers` method retrieves all of sell offers
for the specified NFToken.
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class NFTSellOffers(Request):
    """
    The `nft_sell_offers` method retrieves all of sell offers
    for the specified NFToken.
    """

    method: RequestMethod = field(default=RequestMethod.NFT_SELL_OFFERS, init=False)
    nft_id: str = REQUIRED  # type: ignore
    """
    The unique identifier of an NFToken.
    The request returns sell offers for this NFToken. This value is required.

    :meta hide-value:
    """



"""Model used in AMMBid transaction."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Type

from xrpl.models.base_model import BaseModel
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AuthAccount(BaseModel):
    """Represents one entry in a list of AuthAccounts used in AMMBid transaction."""

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    @classmethod
    def is_dict_of_model(cls: Type[AuthAccount], dictionary: Dict[str, Any]) -> bool:
        """
        Returns True if the input dictionary was derived by the `to_dict`
        method of an instance of this class. In other words, True if this is
        a dictionary representation of an instance of this class.

        NOTE: does not account for model inheritance, IE will only return True
        if dictionary represents an instance of this class, but not if
        dictionary represents an instance of a subclass of this class.

        Args:
            dictionary: The dictionary to check.

        Returns:
            True if dictionary is a dict representation of an instance of this
            class.
        """
        return (
            isinstance(dictionary, dict)
            and "auth_account" in dictionary
            and super().is_dict_of_model(dictionary["auth_account"])
        )

    @classmethod
    def from_dict(cls: Type[AuthAccount], value: Dict[str, Any]) -> AuthAccount:
        """
        Construct a new AuthAccount from a dictionary of parameters.

        Args:
            value: The value to construct the AuthAccount from.

        Returns:
            A new AuthAccount object, constructed using the given parameters.
        """
        if len(value) == 1 and "auth_account" in value:
            return super(AuthAccount, cls).from_dict(value["auth_account"])
        return super(AuthAccount, cls).from_dict(value)

    def to_dict(self: AuthAccount) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a AuthAccount.

        Returns:
            The dictionary representation of a AuthAccount.
        """
        return {"auth_account": super().to_dict()}



"""A sync client for interacting with the rippled WebSocket API."""
from __future__ import annotations

from asyncio import AbstractEventLoop, new_event_loop, run_coroutine_threadsafe
from concurrent.futures import CancelledError, TimeoutError
from threading import Thread
from types import TracebackType
from typing import Any, Dict, Iterator, Optional, Type, Union, cast

from xrpl.asyncio.clients.exceptions import XRPLWebsocketException
from xrpl.asyncio.clients.websocket_base import WebsocketBase
from xrpl.clients.sync_client import SyncClient
from xrpl.models.requests.request import Request
from xrpl.models.response import Response


class WebsocketClient(SyncClient, WebsocketBase):
    """
    A sync client for interacting with the rippled WebSocket API.

    Instead of calling ``open`` and ``close`` yourself, you
    can use a context like so::

        with WebsocketClient(url) as client:
            # inside the context the client is open
        # after exiting the context, the client is closed

    Doing this will open and close the client for you and is
    preferred.

    NOTE: if you are not using subscriptions or other WebSocket-only
    features of rippled, you may not need to do anything other than
    open the client and make requests::

        from xrpl.clients import WebsocketClient
        from xrpl.ledger import get_fee
        from xrpl.models import Fee


        with WebsocketClient(url) as client:
            # using helper functions
            print(get_fee(client))

            # using raw requests yourself
            print(client.request(Fee())

    However, if you are using some functionality that makes use of
    subscriptions or other "websocket-y" things, you can iterate over
    the client like so to read incoming messages::

        with WebsocketClient(url) as client:
            # inside the context the client is open
            for message in client:
                # do something with a message
        # after exiting the context, the client is closed

    NOTE: doing the above will cause the client to listen for
    messages indefinitely. For this reason, ``WebsocketClient``
    takes an optional ``timeout`` parameter which will stop
    iterating on messages if none are received in that timeframe.
    Generally, if you have complex needs with python, xrpl, and
    websockets, you should consider using the ``asyncio`` support
    provided by this library and the ``xrpl.asyncio.clients.AsyncWebsocketClient``
    instead.
    """

    def __init__(
        self: WebsocketClient, url: str, timeout: Optional[Union[int, float]] = None
    ) -> None:
        """
        Constructs a WebsocketClient.

        Arguments:
            url: The URL of the rippled node to submit requests to.
            timeout: Maximum seconds to wait for a new message when
                iterating. A value of 0 or None will result in no limit.
                If this limit is met, iteration will stop.
        """
        self.timeout = timeout
        self._loop: Optional[AbstractEventLoop] = None
        self._thread: Optional[Thread] = None
        super().__init__(url)

    def is_open(self: WebsocketClient) -> bool:
        """
        Returns whether the client is currently open.

        Returns:
            True if the client is currently open, False otherwise.
        """
        return self._loop is not None and self._thread is not None and super().is_open()

    def open(self: WebsocketClient) -> None:
        """Connects the client to the Web Socket API at the given URL."""
        if self.is_open():
            return

        # make a new asyncio event loop
        self._loop = new_event_loop()

        # create and start a thread to run that event loop
        self._thread = Thread(
            target=self._loop.run_forever,
            daemon=True,
        )
        self._thread.start()

        # run WebsocketBase._do_open on the event loop of the child thread and
        # wait for it to finish
        run_coroutine_threadsafe(self._do_open(), self._loop).result()

    def close(self: WebsocketClient) -> None:
        """Closes the connection."""
        if not self.is_open():
            return

        # run WebsocketBase._do_close on the event loop of the child thread and
        # wait for it to finish
        run_coroutine_threadsafe(
            self._do_close(), cast(AbstractEventLoop, self._loop)
        ).result()

        # request the child thread to stop the loop and wait for it to
        # terminate
        cast(AbstractEventLoop, self._loop).call_soon_threadsafe(
            cast(AbstractEventLoop, self._loop).stop
        )
        cast(Thread, self._thread).join()

        # close the stopped loop
        cast(AbstractEventLoop, self._loop).close()

        # clear state
        self._loop = None
        self._thread = None

    def __enter__(self: WebsocketClient) -> WebsocketClient:
        """
        Enters a context after opening itself.

        Returns:
            The opened client.
        """
        self.open()
        return self

    def __exit__(
        self: WebsocketClient,
        _exc_type: Type[BaseException],
        _exc_val: BaseException,
        _trace: TracebackType,
    ) -> None:
        """Exits a context after closing itself."""
        self.close()

    def __iter__(self: WebsocketClient) -> Iterator[Dict[str, Any]]:
        """
        Iterate on received messages. This iterator will block until
        a message is received. If no message is received within
        `self.timeout` seconds then the iterator will exit. If
        `self.timeout` is `None` or `0` then the iterator will block
        indefinetly for the next messsage.
        """
        while self.is_open():
            future = run_coroutine_threadsafe(
                self._do_pop_message(), cast(AbstractEventLoop, self._loop)
            )
            try:
                yield future.result(self.timeout)
            except TimeoutError:
                # in this case, the future reached its timeout. we can safely
                # cancel and stop listening
                future.cancel()
                break
            except CancelledError:
                # in this case, the future was cancelled by someone else. we
                # stop listening but don't need to cancel it
                break

    def send(self: WebsocketClient, request: Request) -> None:
        """
        Submit the request represented by the request to the
        rippled node specified by this client's URL. Unlike ``request``,
        ``send`` does not wait for this request's response. In many cases
        it may be more convenient to use ``request``.

        Arguments:
            request: A Request object representing information about a rippled request.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketClient is not open.
        """
        if not self.is_open():
            raise XRPLWebsocketException("Websocket is not open")
        run_coroutine_threadsafe(
            self._do_send(request), cast(AbstractEventLoop, self._loop)
        ).result()

    async def _request_impl(self: WebsocketClient, request: Request) -> Response:
        """
        ``_request_impl`` implementation for sync websockets that ensures the
        ``WebsocketBase._do_request_impl`` implementation is run on the other thread.

        Arguments:
            request: An object representing information about a rippled request.

        Returns:
            The response from the server, as a Response object.

        Raises:
            XRPLWebsocketException: If there is already an open request by the
                request's ID, or if this WebsocketClient is not open.

        :meta private:
        """
        if not self.is_open():
            raise XRPLWebsocketException("Websocket is not open")

        # it's unusual to write an async function that has no `await` and also
        # has no `async with` or `async for` but in this case that's
        # exactly what we want. the reason we need this is that the helper
        # functions all expect async functions, but since this
        # is a sync client we want to completely block until the request is
        # complete. also, `asyncio.run_coroutine_threadsafe` returns a
        # concurrent.futures.Future which is not awaitable.
        #
        # when this is run via `await client._request_impl`, it will
        # completely block the main thread until completed,
        # just as if it were not async.
        return run_coroutine_threadsafe(
            self._do_request_impl(request),
            cast(AbstractEventLoop, self._loop),
        ).result()



"""All transacion flags and utils to build a list of ints from a FlagInterface"""

from typing import Dict, List, Union

from typing_extensions import TypedDict

from xrpl.models.exceptions import XRPLModelException
from xrpl.models.transactions.types.pseudo_transaction_type import PseudoTransactionType
from xrpl.models.transactions.types.transaction_type import TransactionType

TX_FLAGS: Dict[str, Dict[str, int]] = {
    "AccountSet": {
        "asf_account_tx_id": 0x00000005,
        "asf_authorized_nftoken_minter": 0x0000000A,
        "asf_default_ripple": 0x00000008,
        "asf_deposit_auth": 0x00000009,
        "asf_disable_master": 0x00000004,
        "asf_disallow_xrp": 0x00000003,
        "asf_global_freeze": 0x00000007,
        "asf_no_freeze": 0x00000006,
        "asf_require_auth": 0x00000002,
        "asf_require_dest": 0x00000001,
    },
    "AMMDeposit": {
        "tf_lp_token": 0x00010000,
        "tf_single_asset": 0x00080000,
        "tf_two_asset": 0x00100000,
        "tf_one_asset_lp_token": 0x00200000,
        "tf_limit_lp_token": 0x00400000,
    },
    "AMMWithdraw": {
        "tf_lp_token": 0x00010000,
        "tf_withdraw_all": 0x00020000,
        "tf_one_asset_withdraw_all": 0x00040000,
        "tf_single_asset": 0x00080000,
        "tf_two_asset": 0x00100000,
        "tf_one_asset_lp_token": 0x00200000,
        "tf_limit_lp_token": 0x00400000,
    },
    "NFTokenCreateOffer": {
        "tf_sell_token": 0x00000001,
    },
    "NFTokenMint": {
        "tf_burnable": 0x00000001,
        "tf_only_xrp": 0x00000002,
        "tf_trustline": 0x00000004,
        "tf_transferable": 0x00000008,
    },
    "OfferCreate": {
        "tf_passive": 0x00010000,
        "tf_immediate_or_cancel": 0x00020000,
        "tf_fill_or_kill": 0x00040000,
        "tf_sell": 0x00080000,
    },
    "Payment": {
        "tf_no_direct_ripple": 0x00010000,
        "tf_partial_payment": 0x00020000,
        "tf_limit_quality": 0x00040000,
    },
    "PaymentChannelClaim": {
        "tf_renew": 0x00010000,
        "tf_close": 0x00020000,
    },
    "TrustSet": {
        "tf_set_auth": 0x00010000,
        "tf_set_no_ripple": 0x00020000,
        "tf_clear_no_ripple": 0x00040000,
        "tf_set_freeze": 0x00100000,
        "tf_clear_freeze": 0x00200000,
    },
    "EnableAmendment": {
        "tf_got_majority": 0x00010000,
        "tf_lost_majority": 0x00020000,
    },
}


class FlagInterface(TypedDict):
    """A TypedDict to define transaction flags by bool."""

    pass


def interface_to_flag_list(
    tx_type: Union[TransactionType, PseudoTransactionType],
    tx_flags: Dict[str, bool],
) -> List[int]:
    """Parse a list of flags expressed as integers from the FlagInterface.

    Args:
        tx_type (Union[TransactionType, PseudoTransactionType]):
            Type of the transaction.
        tx_flags (dict):
            FlagInterface

    Returns:
        List[int]:
            A list of flags expressed as integers.
    """
    if tx_type not in TX_FLAGS:  # if transaction types has no flags
        return [0]
    flags = TX_FLAGS[tx_type]
    flag_list = []
    for flag, num in flags.items():
        if flag in tx_flags:  # if flag was defined
            if tx_flags[flag]:  # if flag was set to True
                flag_list.append(num)
            else:
                flag_list.append(0)  # if flag was set to False
        else:  # if flag was not defined
            flag_list.append(0)
    return flag_list


def check_false_flag_definition(
    tx_type: Union[TransactionType, PseudoTransactionType],
    tx_flags: Union[Dict[str, bool], List[int]],
) -> None:
    """Check the flags were set correctly if not defined as integer.

    Args:
        tx_type (Union[TransactionType, PseudoTransactionType]):
            Type of the transaction.
        tx_flags (Iterable):
            FlagInterface

    Retruns:
        None

    Raises:
        XRPLModelException: Flags were not set correctly.
    """
    try:
        if isinstance(tx_flags, dict):
            interface_flags: Dict[str, bool] = tx_flags
            assert all(
                [
                    all(
                        (
                            isinstance(flag, str),
                            isinstance(set_flag, bool),
                        )
                    )
                    for flag, set_flag in interface_flags.items()
                ]
            )
        else:  # List[int]
            list_flags: List[int] = tx_flags
            assert all(isinstance(flag, int) for flag in list_flags)
    except AssertionError:
        msg = f"""
False flag definition: Please define flags either by setting bools using
`{tx_type}FlagInterface` or by using `{tx_type}Flag`.
Do not put the FlagInterface in a list or mix them.
""".strip()
        raise XRPLModelException(msg)



"""Sidechain-related helper util functions."""

from xrpl.constants import XRPLException
from xrpl.models import Memo, Payment
from xrpl.utils.str_conversions import str_to_hex


def create_cross_chain_payment(payment: Payment, dest_account: str) -> Payment:
    """
    Creates a cross-chain payment transaction.

    Args:
        payment: The initial payment transaction. If the transaction is signed, then
            it will need to be re-signed. There must be no more than 2 memos, since the
            first memo is used for the sidechain destination account. The destination
            must be the sidechain's door account.
        dest_account: The destination account on the sidechain.

    Returns:
        A cross-chain payment transaction, where the mainchain door account is the
            `Destination` and the destination account on the sidechain is encoded in
            the memos.

    Raises:
        XRPLException: If there are more than 2 memos.
    """
    dest_account_memo = Memo(memo_data=str_to_hex(dest_account))

    if payment.memos is None:
        memos = []
    else:
        memos = payment.memos
    if memos is not None and len(memos) > 2:
        raise XRPLException(
            "Cannot have more than 2 memos in a cross-chain transaction."
        )
    new_memos = [dest_account_memo] + memos

    payment_dict = payment.to_dict()
    payment_dict["memos"] = new_memos
    if "txn_signature" in payment_dict:
        del payment_dict["txn_signature"]
    return Payment.from_dict(payment_dict)



"""
The base class for all network request types.
Represents fields common to all request types.
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Type, TypeVar, Union, cast

import xrpl.models.requests  # bare import to get around circular dependency
from xrpl.models.base_model import BaseModel
from xrpl.models.exceptions import XRPLModelException
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


class RequestMethod(str, Enum):
    """Represents the different options for the ``method`` field in a request."""

    # account methods
    ACCOUNT_CHANNELS = "account_channels"
    ACCOUNT_CURRENCIES = "account_currencies"
    ACCOUNT_INFO = "account_info"
    ACCOUNT_LINES = "account_lines"
    ACCOUNT_NFTS = "account_nfts"
    ACCOUNT_OBJECTS = "account_objects"
    ACCOUNT_OFFERS = "account_offers"
    ACCOUNT_TX = "account_tx"
    GATEWAY_BALANCES = "gateway_balances"
    NO_RIPPLE_CHECK = "noripple_check"

    # transaction methods
    SIGN = "sign"
    SIGN_FOR = "sign_for"
    SUBMIT = "submit"
    SUBMIT_MULTISIGNED = "submit_multisigned"
    TRANSACTION_ENTRY = "transaction_entry"
    TX = "tx"

    # channel methods
    CHANNEL_AUTHORIZE = "channel_authorize"
    CHANNEL_VERIFY = "channel_verify"

    # path methods
    BOOK_OFFERS = "book_offers"
    DEPOSIT_AUTHORIZED = "deposit_authorized"
    PATH_FIND = "path_find"
    RIPPLE_PATH_FIND = "ripple_path_find"

    # ledger methods
    LEDGER = "ledger"
    LEDGER_CLOSED = "ledger_closed"
    LEDGER_CURRENT = "ledger_current"
    LEDGER_DATA = "ledger_data"
    LEDGER_ENTRY = "ledger_entry"

    # NFT methods
    NFT_BUY_OFFERS = "nft_buy_offers"
    NFT_SELL_OFFERS = "nft_sell_offers"

    # subscription methods
    SUBSCRIBE = "subscribe"
    UNSUBSCRIBE = "unsubscribe"

    # server info methods
    FEE = "fee"
    MANIFEST = "manifest"
    SERVER_INFO = "server_info"
    SERVER_STATE = "server_state"

    # utility methods
    PING = "ping"
    RANDOM = "random"

    # sidechain methods
    FEDERATOR_INFO = "federator_info"

    # amm methods
    AMM_INFO = "amm_info"

    # generic unknown/unsupported request
    # (there is no XRPL analog, this model is specific to xrpl-py)
    GENERIC_REQUEST = "zzgeneric_request"


R = TypeVar("R", bound="Request")


@require_kwargs_on_init
@dataclass(frozen=True)
class Request(BaseModel):
    """
    The base class for all network request types.
    Represents fields common to all request types.
    """

    method: RequestMethod = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    id: Optional[Union[str, int]] = None

    @classmethod
    def from_dict(cls: Type[R], value: Dict[str, Any]) -> R:
        """
        Construct a new Request from a dictionary of parameters.

        Args:
            value: The value to construct the Request from.

        Returns:
            A new Request object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        if cls.__name__ == "Request":
            if "method" not in value:
                raise XRPLModelException("Request does not include method.")
            correct_type = cls.get_method(value["method"])
            return correct_type.from_dict(value)  # type: ignore

        if "method" in value:
            method = value["method"]
            if (
                cls.get_method(method).__name__ != cls.__name__
                and not (
                    method == "submit"
                    and cls.__name__ in ("SignAndSubmit", "SubmitOnly")
                )
                and not cls.__name__ == "GenericRequest"
            ):
                raise XRPLModelException(
                    f"Using wrong constructor: using {cls.__name__} constructor "
                    f"with Request method {method}."
                )
            value = {**value}
            del value["method"]

        return super(Request, cls).from_dict(value)

    @classmethod
    def get_method(cls: Type[Request], method: str) -> Type[Request]:
        """
        Returns the correct request method based on the string name.

        Args:
            method: The String name of the Request object.

        Returns:
            The request class with the given name. If the request doesn't exist, then
            it will return a `GenericRequest`.
        """
        # special case for NoRippleCheck and NFT methods
        if method == RequestMethod.NO_RIPPLE_CHECK:
            return xrpl.models.requests.NoRippleCheck
        if method == RequestMethod.ACCOUNT_NFTS:
            return xrpl.models.requests.AccountNFTs
        if method == RequestMethod.NFT_BUY_OFFERS:
            return xrpl.models.requests.NFTBuyOffers
        if method == RequestMethod.NFT_SELL_OFFERS:
            return xrpl.models.requests.NFTSellOffers

        parsed_name = "".join([word.capitalize() for word in method.split("_")])
        if parsed_name in xrpl.models.requests.__all__:
            return cast(Type[Request], getattr(xrpl.models.requests, parsed_name))
        return xrpl.models.requests.GenericRequest

    def to_dict(self: Request) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a Request.

        Returns:
            The dictionary representation of a Request.
        """
        # we need to override this because method is using ``field``
        # which will not include the value in the object's __dict__
        return {**super().to_dict(), "method": self.method.value}



"""
The XRP Ledger has two kinds of money: XRP, and issued
currencies. Both types have high precision, although their
formats are different.
"""
from xrpl.models.currencies.currency import Currency
from xrpl.models.currencies.issued_currency import IssuedCurrency
from xrpl.models.currencies.xrp import XRP

__all__ = [
    "Currency",
    "IssuedCurrency",
    "XRP",
]



"""Class for serializing and deserializing a 32-bit UInt.
See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
"""
from __future__ import annotations

from typing import Optional, Type, Union

from typing_extensions import Final

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.uint import UInt

_WIDTH: Final[int] = 4  # 32 / 8


class UInt32(UInt):
    """
    Class for serializing and deserializing a 32-bit UInt.
    See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
    """

    def __init__(self: UInt32, buffer: bytes = bytes(_WIDTH)) -> None:
        """Construct a new UInt32 type from a ``bytes`` value."""
        super().__init__(buffer)

    @classmethod
    def from_parser(
        cls: Type[UInt32], parser: BinaryParser, _length_hint: Optional[int] = None
    ) -> UInt32:
        """
        Construct a new UInt32 type from a BinaryParser.

        Args:
            parser: A BinaryParser to construct a UInt32 from.

        Returns:
            The UInt32 constructed from parser.
        """
        return cls(parser.read(_WIDTH))

    @classmethod
    def from_value(cls: Type[UInt32], value: Union[str, int]) -> UInt32:
        """
        Construct a new UInt32 type from a number.

        Args:
            value: The number to construct a UInt32 from.

        Returns:
            The UInt32 constructed from value.

        Raises:
            XRPLBinaryCodecException: If a UInt32 could not be constructed from value.
        """
        if not isinstance(value, (str, int)):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a UInt32: expected str or int,"
                " received {value.__class__.__name__}."
            )

        if isinstance(value, int):
            value_bytes = (value).to_bytes(_WIDTH, byteorder="big", signed=False)
            return cls(value_bytes)

        if isinstance(value, str) and value.isdigit():
            value_bytes = (int(value)).to_bytes(_WIDTH, byteorder="big", signed=False)
            return cls(value_bytes)

        raise XRPLBinaryCodecException("Cannot construct UInt32 from given value")



"""Parse final balances of every account involved in the given transaction."""

from decimal import Decimal
from typing import List, Optional

from xrpl.models import TransactionMetadata
from xrpl.utils.txn_parser.utils import (
    AccountBalances,
    NormalizedNode,
    derive_account_balances,
    get_value,
)


def get_final_balances(metadata: TransactionMetadata) -> List[AccountBalances]:
    """
    Parse all final balances from a transaction's metadata.

    Args:
        metadata: Transactions metadata.

    Returns:
        All final balances caused by a transaction.
        The final balances are grouped by the affected account addresses.
    """
    return derive_account_balances(metadata, _compute_final_balance)


def _compute_final_balance(node: NormalizedNode) -> Optional[Decimal]:
    """
    Get the final balance from a node.

    Args:
        node: The affected node.

    Returns:
        The final balance.
    """
    value: Optional[Decimal] = None
    new_fields = node.get("NewFields")
    final_fields = node.get("FinalFields")
    if new_fields is not None:
        balance = new_fields.get("Balance")
        if balance is not None:
            value = get_value(balance)
    elif final_fields is not None:
        balance = final_fields.get("Balance")
        if balance is not None:
            value = get_value(balance)
    if value is None or value == Decimal(0):
        return None
    return value



"""Model for EscrowCreate transaction type."""
from __future__ import annotations  # Requires Python 3.7+

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.amounts import Amount
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class EscrowCreate(Transaction):
    """
    Represents an `EscrowCreate <https://xrpl.org/escrowcreate.html>`_
    transaction, which locks up amount until a specific time or condition is met.
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    Amount to deduct from the sender's balance and escrow. Once escrowed, the
    amount can either go to the Destination address (after the FinishAfter time)
    or returned to the sender (after the CancelAfter time). This field is required.

    :meta hide-value:
    """

    destination: str = REQUIRED  # type: ignore
    """
    The address that should receive the escrowed amount when the time or
    condition is met. This field is required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    An arbitrary `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ that
    identifies the reason for the Escrow, or a hosted recipient to pay.
    """

    cancel_after: Optional[int] = None
    """
    The time, in seconds since the Ripple Epoch, when this escrow expires.
    This value is immutable; the funds can only be returned the sender after
    this time.
    """

    finish_after: Optional[int] = None
    """
    The time, in seconds since the Ripple Epoch, when the escrowed amount can
    be released to the recipient. This value is immutable; the funds cannot
    move until this time is reached.
    """

    condition: Optional[str] = None
    """
    Hex value representing a `PREIMAGE-SHA-256 crypto-condition
    <https://tools.ietf.org/html/draft-thomas-crypto-conditions-04#section-8.1.>`_
    The funds can only be delivered to the recipient if this condition is
    fulfilled.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.ESCROW_CREATE,
        init=False,
    )

    def _get_errors(self: EscrowCreate) -> Dict[str, str]:
        errors = super()._get_errors()
        if (
            self.cancel_after is not None
            and self.finish_after is not None
            and self.finish_after >= self.cancel_after
        ):
            errors[
                "EscrowCreate"
            ] = "The finish_after time must be before the cancel_after time."

        return errors



"""A collection of serialization information about a specific field type."""
from __future__ import annotations  # Requires Python 3.7+

from typing import TYPE_CHECKING, Dict, Type

from xrpl.core.binarycodec.definitions.field_header import FieldHeader
from xrpl.core.binarycodec.definitions.field_info import FieldInfo

if TYPE_CHECKING:
    # To prevent a circular dependency.
    from xrpl.core.binarycodec.types.serialized_type import SerializedType


def _get_type_by_name(name: str) -> Type[SerializedType]:
    """
    Convert the string name of a class to the class object itself.

    Args:
        name: the name of the class.

    Returns:
        The corresponding class object.
    """
    import xrpl.core.binarycodec.types as types

    type_map: Dict[str, Type[SerializedType]] = {
        name: object_type
        for (name, object_type) in types.__dict__.items()
        if name in types.__all__
    }

    return type_map[name]


class FieldInstance:
    """A collection of serialization information about a specific field type."""

    def __init__(
        self: FieldInstance,
        field_info: FieldInfo,
        field_name: str,
        field_header: FieldHeader,
    ) -> None:
        """
        Construct a FieldInstance.

        :param field_info: The field's serialization info from definitions.json.
        :param field_name: The field's string name.
        :param field_header: A FieldHeader object with the type_code and field_code.
        """
        self.nth = field_info.nth
        self.is_variable_length_encoded = field_info.is_variable_length_encoded
        self.is_serialized = field_info.is_serialized
        self.is_signing = field_info.is_signing_field
        self.type = field_info.type
        self.name = field_name
        self.header = field_header
        self.ordinal = self.header.type_code << 16 | self.nth
        self.associated_type = _get_type_by_name(self.type)



"""
The submit_multisigned command applies a multi-signed transaction and sends it to the
network to be included in future ledgers. (You can also submit multi-signed
transactions in binary form using the submit command in submit-only mode.)

This command requires the MultiSign amendment to be enabled.

`See submit_multisigned <https://xrpl.org/submit_multisigned.html>`_
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Type

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class SubmitMultisigned(Request):
    """
    The submit_multisigned command applies a multi-signed transaction and sends it to
    the network to be included in future ledgers. (You can also submit multi-signed
    transactions in binary form using the submit command in submit-only mode.)

    This command requires the MultiSign amendment to be enabled.

    `See submit_multisigned <https://xrpl.org/submit_multisigned.html>`_
    """

    method: RequestMethod = field(default=RequestMethod.SUBMIT_MULTISIGNED, init=False)
    tx_json: Transaction = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    fail_hard: bool = False

    @classmethod
    def from_dict(
        cls: Type[SubmitMultisigned], value: Dict[str, Any]
    ) -> SubmitMultisigned:
        """
        Construct a new SubmitMultisigned object from a dictionary of parameters.

        Args:
            value: The value to construct the SubmitMultisigned from.

        Returns:
            A new SubmitMultisigned object, constructed using the given parameters.
        """
        fixed_value = {**value}
        if "TransactionType" in fixed_value["tx_json"]:  # xrpl format
            fixed_value["tx_json"] = Transaction.from_xrpl(fixed_value["tx_json"])
        return super(SubmitMultisigned, cls).from_dict(fixed_value)

    def to_dict(self: SubmitMultisigned) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a SubmitMultisigned object.

        Returns:
            The dictionary representation of a SubmitMultisigned object.
        """
        xrpl_dict = super().to_dict()
        xrpl_dict["tx_json"] = self.tx_json.to_xrpl()
        return xrpl_dict



"""
This request calculates the total balances issued by a given account, optionally
excluding amounts held by operational addresses.

`See gateway_balances <https://xrpl.org/gateway_balances.html>`_
"""
from dataclasses import dataclass, field
from typing import List, Optional, Union

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class GatewayBalances(Request):
    """
    This request calculates the total balances issued by a given account, optionally
    excluding amounts held by operational addresses.

    `See gateway_balances <https://xrpl.org/gateway_balances.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    ledger_hash: Optional[str] = None
    ledger_index: Optional[Union[str, int]] = None
    method: RequestMethod = field(default=RequestMethod.GATEWAY_BALANCES, init=False)
    strict: bool = False
    hotwallet: Optional[Union[str, List[str]]] = None



"""The information needed to control an XRPL account."""

from __future__ import annotations

from typing import Optional, Type

from xrpl.constants import CryptoAlgorithm
from xrpl.core.addresscodec import classic_address_to_xaddress
from xrpl.core.keypairs import derive_classic_address, derive_keypair, generate_seed


class Wallet:
    """
    The cryptographic keys needed to control an XRP Ledger account. See
    `Cryptographic Keys <https://xrpl.org/cryptographic-keys.html>`_ for
    details.
    """

    def __init__(
        self: Wallet,
        seed: str,
        sequence: int,
        *,
        algorithm: Optional[CryptoAlgorithm] = None,
    ) -> None:
        """
        Generate a new Wallet.

        Args:
            seed: The seed from which the public and private keys are derived.
            sequence: The next sequence number for the account.
            algorithm: The algorithm used to encode the keys. Inferred from the seed if
                not included.
        """
        self.seed = seed
        """
        The core value that is used to derive all other information about
        this wallet. MUST be kept secret!
        """

        pk, sk = derive_keypair(self.seed, algorithm=algorithm)
        self.algorithm = algorithm
        self.public_key = pk
        """
        The public key that is used to identify this wallet's signatures, as
        a hexadecimal string.
        """

        self.private_key = sk
        """
        The private key that is used to create signatures, as a hexadecimal
        string. MUST be kept secret!
        """

        self.classic_address = derive_classic_address(self.public_key)
        """The address that publicly identifies this wallet, as a base58 string."""

        self.sequence = sequence
        """
        The next available sequence number to use for transactions from this
        wallet.
        Must be updated by the user. Increments on the ledger with every successful
        transaction submission, and stays the same with every failed transaction
        submission.
        """

    @classmethod
    def create(
        cls: Type[Wallet], crypto_algorithm: CryptoAlgorithm = CryptoAlgorithm.ED25519
    ) -> Wallet:
        """
        Generates a new seed and Wallet.

        Args:
            crypto_algorithm: The key-generation algorithm to use when generating the
                seed. The default is Ed25519.

        Returns:
            The wallet that is generated from the given seed.
        """
        seed = generate_seed(algorithm=crypto_algorithm)
        return cls(seed, sequence=0, algorithm=crypto_algorithm)

    def get_xaddress(
        self: Wallet, *, tag: Optional[int] = None, is_test: bool = False
    ) -> str:
        """
        Returns the X-Address of the Wallet's account.

        Args:
            tag: the destination tag of the address. Defaults to `None`.
            is_test: whether the address corresponds to an address on the test network.

        Returns:
            The X-Address of the Wallet's account.
        """
        return classic_address_to_xaddress(self.classic_address, tag, is_test)

    def __str__(self: Wallet) -> str:
        """
        Returns a string representation of a Wallet.

        Returns:
            A string representation of a Wallet.
        """
        return "\n".join(
            [
                f"public_key: {self.public_key}",
                "private_key: -HIDDEN-",
                f"classic_address: {self.classic_address}",
            ]
        )



"""An async client for interacting with the rippled JSON RPC."""
from xrpl.asyncio.clients.async_client import AsyncClient
from xrpl.asyncio.clients.json_rpc_base import JsonRpcBase


class AsyncJsonRpcClient(AsyncClient, JsonRpcBase):
    """An async client for interacting with the rippled JSON RPC."""

    pass



"""High-level methods to obtain information about account transaction history."""
from typing import Any, Dict, List, cast

from deprecated.sphinx import deprecated

from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.core.addresscodec import is_valid_xaddress, xaddress_to_classic_address
from xrpl.models.requests import AccountTx
from xrpl.models.response import Response


async def get_latest_transaction(account: str, client: Client) -> Response:
    """
    Fetches the most recent transaction on the ledger associated with an account.

    Args:
        account: the account to query.
        client: the network client used to communicate with a rippled node.

    Returns:
        The Response object containing the transaction info.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    # max == -1 means that it's the most recent validated ledger version
    if is_valid_xaddress(account):
        account, _, _ = xaddress_to_classic_address(account)
    response = await client._request_impl(
        AccountTx(account=account, ledger_index_max=-1, limit=1)
    )
    if not response.is_successful():
        raise XRPLRequestFailureException(response.result)
    return response


@deprecated(
    reason="Sending an AccountTx request directly allows you to page through all "
    "results and is just as easy to use.",
    version="1.6.0",
)
async def get_account_transactions(
    address: str,
    client: Client,
) -> List[Dict[str, Any]]:
    """
    Query the ledger for a list of transactions that involved a given account.
    To access more than just the first page of results, use the :class:`AccountTx`
    request directly.

    Args:
        address: the account to query.
        client: the network client used to make network calls.

    Returns:
        The most recent transaction history for the address. For the full history,
        page through the :class:`AccountTx` request directly.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    if is_valid_xaddress(address):
        address, _, _ = xaddress_to_classic_address(address)
    request = AccountTx(account=address)
    response = await client._request_impl(request)
    if not response.is_successful():
        raise XRPLRequestFailureException(response.result)
    return cast(List[Dict[str, Any]], response.result["transactions"])


@deprecated(
    reason="Sending an AccountTx request directly and filtering for payments allows "
    "you to page through all results and is just as easy to use.",
    version="1.8.0",
)
async def get_account_payment_transactions(
    address: str,
    client: Client,
) -> List[Dict[str, Any]]:
    """
    Query the ledger for a list of payment transactions that involved a given account.
    To access more than just the first page of results, use the :class:`AccountTx`
    request directly then filter for transactions with a "Payment" TransactionType.

    Args:
        address: the account to query.
        client: the network client used to make network calls.

    Returns:
        The first page of payment transaction history for the address. For the full
        history, page through the :class:`AccountTx` request directly.
    """
    all_transactions = await get_account_transactions(address, client)
    return [tx for tx in all_transactions if tx["tx"]["TransactionType"] == "Payment"]



"""Model for a XChainModifyBridge transaction type."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from xrpl.models.currencies import XRP
from xrpl.models.flags import FlagInterface
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init
from xrpl.models.xchain_bridge import XChainBridge


class XChainModifyBridgeFlag(int, Enum):
    """
    Transactions of the XChainModifyBridge type support additional values in the Flags
    field. This enum represents those options.
    """

    TF_CLEAR_ACCOUNT_CREATE_AMOUNT = 0x00010000


class XChainModifyBridgeFlagInterface(FlagInterface):
    """
    Transactions of the XChainModifyBridge type support additional values in the Flags
    field. This TypedDict represents those options.
    """

    TF_CLEAR_ACCOUNT_CREATE_AMOUNT: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainModifyBridge(Transaction):
    """Represents a XChainModifyBridge transaction."""

    xchain_bridge: XChainBridge = REQUIRED  # type: ignore

    signature_reward: Optional[str] = None

    min_account_create_amount: Optional[str] = None

    transaction_type: TransactionType = field(
        default=TransactionType.XCHAIN_MODIFY_BRIDGE,
        init=False,
    )

    def _get_errors(self: XChainModifyBridge) -> Dict[str, str]:
        errors = super()._get_errors()

        bridge = self.xchain_bridge

        if self.signature_reward is None and self.min_account_create_amount is None:
            errors[
                "xchain_modify_bridge"
            ] = "Must either change signature_reward or min_account_create_amount."

        if self.account not in [bridge.locking_chain_door, bridge.issuing_chain_door]:
            errors[
                "account"
            ] = "Account must be either locking chain door or issuing chain door."

        if self.signature_reward is not None and not self.signature_reward.isnumeric():
            errors["signature_reward"] = "`signature_reward` must be numeric."

        if (
            self.min_account_create_amount is not None
            and bridge.locking_chain_issue != XRP()
        ):
            errors[
                "min_account_create_amount"
            ] = "Cannot have MinAccountCreateAmount if bridge is IOU-IOU."

        if (
            self.min_account_create_amount is not None
            and not self.min_account_create_amount.isnumeric()
        ):
            errors[
                "min_account_create_amount_value"
            ] = "`min_account_create_amount` must be numeric."

        return errors



"""Model for DepositPreauth transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class DepositPreauth(Transaction):
    """
    Represents a `DepositPreauth <https://xrpl.org/depositpreauth.html>`_
    transaction, which gives another account pre-approval to deliver payments to
    the sender of this transaction, if this account is using
    `Deposit Authorization <https://xrpl.org/depositauth.html>`_.
    """

    authorize: Optional[str] = None
    """
    Grant preauthorization to this address. You must provide this OR
    ``unauthorize`` but not both.
    """

    unauthorize: Optional[str] = None
    """
    Revoke preauthorization from this address. You must provide this OR
    ``authorize`` but not both.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.DEPOSIT_PREAUTH,
        init=False,
    )

    def _get_errors(self: DepositPreauth) -> Dict[str, str]:
        errors = super()._get_errors()
        if self.authorize and self.unauthorize:
            errors[
                "DepositPreauth"
            ] = "One of authorize and unauthorize must be set, not both."

        if not self.authorize and not self.unauthorize:
            errors["DepositPreauth"] = "One of authorize and unauthorize must be set."

        return errors



"""
The ledger_closed method returns the unique
identifiers of the most recently closed ledger.
(This ledger is not necessarily validated and
immutable yet.)
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.required import REQUIRED
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class LedgerClosed(Request):
    """
    The ledger_closed method returns the unique
    identifiers of the most recently closed ledger.
    (This ledger is not necessarily validated and
    immutable yet.)
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER_CLOSED, init=False)
    ledger_hash: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    ledger_index: int = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """



"""High-level methods that fetch transaction information from the XRP Ledger."""

import asyncio
from typing import Optional

from deprecated.sphinx import deprecated

from xrpl.asyncio.transaction import ledger
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response


@deprecated(
    reason="Sending a Tx request directly is just as easy to use.",
    version="1.8.0",
)
def get_transaction_from_hash(
    tx_hash: str,
    client: SyncClient,
    binary: bool = False,
    min_ledger: Optional[int] = None,
    max_ledger: Optional[int] = None,
) -> Response:
    """
    Given a transaction hash, fetch the corresponding transaction from the ledger.

    Args:
        tx_hash: the transaction hash.
        client: the network client used to communicate with a rippled node.
        binary: If true, return transaction data and metadata as binary
            serialized to hexadecimal strings. If false, return transaction data and
            metadata as JSON. The default is false.
        min_ledger: Use this with max_ledger to specify a range of up to
            1000 ledger indexes, starting with this ledger (inclusive). If the server
            cannot find the transaction, it confirms whether it was able to search all
            the ledgers in this range.
        max_ledger: Use this with min_ledger to specify a range of up to
            1000 ledger indexes, ending with this ledger (inclusive). If the server
            cannot find the transaction, it confirms whether it was able to search
            all the ledgers in the requested range.

    Returns:
        The Response object containing the transaction info.

    Raises:
        XRPLRequestFailureException: if the transaction fails.
    """
    return asyncio.run(
        ledger.get_transaction_from_hash(
            tx_hash,
            client,
            binary,
            min_ledger,
            max_ledger,
        )
    )



"""
Class for serializing and deserializing a 64-bit UInt.
See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
"""
from __future__ import annotations

import re
from typing import Optional, Pattern, Type, Union

from typing_extensions import Final

from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.uint import UInt

_WIDTH: Final[int] = 8  # 64 / 8

_HEX_REGEX: Final[Pattern[str]] = re.compile("[a-fA-F0-9]{1,16}")


class UInt64(UInt):
    """
    Class for serializing and deserializing a 64-bit UInt.
    See `UInt Fields <https://xrpl.org/serialization.html#uint-fields>`_
    """

    def __init__(self: UInt64, buffer: bytes = bytes(_WIDTH)) -> None:
        """Construct a new UInt64 type from a ``bytes`` value."""
        super().__init__(buffer)

    @classmethod
    def from_parser(
        cls: Type[UInt64], parser: BinaryParser, _length_hint: Optional[int] = None
    ) -> UInt64:
        """
        Construct a new UInt64 type from a BinaryParser.

        Args:
            parser: The BinaryParser to construct a UInt64 from.

        Returns:
            The UInt64 constructed from parser.
        """
        return cls(parser.read(_WIDTH))

    @classmethod
    def from_value(cls: Type[UInt64], value: Union[str, int]) -> UInt64:
        """
        Construct a new UInt64 type from a number.

        Args:
            value: The number to construct a UInt64 from.

        Returns:
            The UInt64 constructed from value.

        Raises:
            XRPLBinaryCodecException: If a UInt64 could not be constructed from value.
        """
        if not isinstance(value, (str, int)):
            raise XRPLBinaryCodecException(
                "Invalid type to construct a UInt64: expected str or int,"
                " received {value.__class__.__name__}."
            )

        if isinstance(value, int):
            if value < 0:
                raise XRPLBinaryCodecException("{value} must be an unsigned integer")
            value_bytes = (value).to_bytes(_WIDTH, byteorder="big", signed=False)
            return cls(value_bytes)

        if isinstance(value, str):
            if not _HEX_REGEX.fullmatch(value):
                raise XRPLBinaryCodecException("{value} is not a valid hex string")
            value = value.rjust(16, "0")
            value_bytes = bytes.fromhex(value)
            return cls(value_bytes)

        raise XRPLBinaryCodecException(
            "Cannot construct UInt64 from given value {value}"
        )

    def to_json(self: UInt64) -> str:
        """
        Convert a UInt64 object to JSON (hex).

        Returns:
            The JSON representation of the UInt64 object.
        """
        return self.buffer.hex().upper()



"""
The ping command returns an acknowledgement, so that
clients can test the connection status and latency.
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class Ping(Request):
    """
    The ping command returns an acknowledgement, so that
    clients can test the connection status and latency.
    """

    method: RequestMethod = field(default=RequestMethod.PING, init=False)



"""Wrapper classes around byte buffers used for serialization and deserialization."""
from xrpl.core.binarycodec.binary_wrappers.binary_parser import BinaryParser
from xrpl.core.binarycodec.binary_wrappers.binary_serializer import BinarySerializer

__all__ = ["BinaryParser", "BinarySerializer"]



"""Codec for serializing and deserializing AccountID fields.
See `AccountID Fields <https://xrpl.org/serialization.html#accountid-fields>`_
"""
from __future__ import annotations  # Requires Python 3.7+

import re
from typing import Optional, Pattern, Type

from typing_extensions import Final

from xrpl.core.addresscodec import (
    decode_classic_address,
    encode_classic_address,
    is_valid_classic_address,
    is_valid_xaddress,
    xaddress_to_classic_address,
)
from xrpl.core.binarycodec import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.hash160 import Hash160

# matches hex-encoded accounts. this happens to be the same format as the
# hex currency encoding, but that might change so we're redefining it here.
_HEX_REGEX: Final[Pattern[str]] = re.compile("[A-F0-9]{40}")


class AccountID(Hash160):
    """Codec for serializing and deserializing AccountID fields.
    See `AccountID Fields <https://xrpl.org/serialization.html#accountid-fields>`_
    """

    LENGTH: Final[int] = 20  # bytes

    def __init__(self: AccountID, buffer: Optional[bytes] = None) -> None:
        """
        Construct an AccountID from given bytes.
        If buffer is not provided, default to 20 zero bytes.
        """
        if buffer is not None and len(buffer) > 0:
            super().__init__(buffer)
        else:
            super().__init__(bytes(self.LENGTH))

    @classmethod
    def from_value(cls: Type[AccountID], value: str) -> AccountID:
        """
        Construct an AccountID from a hex string or a base58 r-Address.

        Args:
            value: The string to construct an AccountID from.

        Returns:
            The AccountID constructed from value.

        Raises:
            XRPLBinaryCodecException: If the supplied value is of the wrong type.
        """
        if not isinstance(value, str):
            raise XRPLBinaryCodecException(
                "Invalid type to construct an AccountID: expected str,"
                f" received {value.__class__.__name__}."
            )

        if value == "":
            return cls()

        # hex-encoded case
        if _HEX_REGEX.fullmatch(value):
            return cls(bytes.fromhex(value))
        # base58 case
        if is_valid_classic_address(value):
            return cls(decode_classic_address(value))
        if is_valid_xaddress(value):
            classic_address, _, _ = xaddress_to_classic_address(value)
            return cls(decode_classic_address(classic_address))

        raise XRPLBinaryCodecException(
            "Invalid value to construct an AccountID: expected valid classic address "
            f"or X-Address, received {value.__class__.__name__}."
        )

    def to_json(self: AccountID) -> str:
        """
        Return the value of this AccountID encoded as a base58 string.

        Returns:
            The JSON representation of the AccountID.
        """
        return encode_classic_address(self.buffer)



"""High-level reliable submission methods with XRPL transactions."""

import asyncio

from xrpl.asyncio.transaction import (
    send_reliable_submission as async_send_reliable_submission,
)
from xrpl.clients.sync_client import SyncClient
from xrpl.models.response import Response
from xrpl.models.transactions.transaction import Transaction


def send_reliable_submission(transaction: Transaction, client: SyncClient) -> Response:
    """
    Submits a transaction and verifies that it has been included in a validated ledger
    (or has errored/will not be included for some reason).

    `See Reliable Transaction Submission
    <https://xrpl.org/reliable-transaction-submission.html>`_

    Note: This cannot be used with a standalone rippled node, because ledgers do not
    close automatically.

    Args:
        transaction: the signed transaction to submit to the ledger. Requires a
            `last_ledger_sequence` param.
        client: the network client used to submit the transaction to a rippled node.

    Returns:
        The response from a validated ledger.
    """
    return asyncio.run(async_send_reliable_submission(transaction, client))



"""Model for SignerListSet transaction type."""
from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Pattern

from typing_extensions import Final

from xrpl.models.nested_model import NestedModel
from xrpl.models.required import REQUIRED
from xrpl.models.transactions.transaction import Transaction
from xrpl.models.transactions.types import TransactionType
from xrpl.models.utils import require_kwargs_on_init

MAX_SIGNER_ENTRIES: Final[int] = 32
"""
Maximum number of signer entries allowed.

:meta private:
"""

HEX_WALLET_LOCATOR_REGEX: Final[Pattern[str]] = re.compile("[A-Fa-f0-9]{64}")
"""
Matches hex-encoded WalletLocator in the format allowed by XRPL.

:meta private:
"""


@require_kwargs_on_init
@dataclass(frozen=True)
class SignerEntry(NestedModel):
    """Represents one entry in a list of multi-signers authorized to an account."""

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    signer_weight: int = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    wallet_locator: Optional[str] = None
    """
    An arbitrary 256-bit (32-byte) field that can be used to identify the signer, which
    may be useful for smart contracts, or for identifying who controls a key in a large
    organization.
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class SignerListSet(Transaction):
    """
    Represents a `SignerListSet <https://xrpl.org/signerlistset.html>`_
    transaction, which creates, replaces, or removes a list of signers that
    can be used to `multi-sign a transaction
    <https://xrpl.org/multi-signing.html>`_.
    """

    signer_quorum: int = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    signer_entries: Optional[List[SignerEntry]] = None
    transaction_type: TransactionType = field(
        default=TransactionType.SIGNER_LIST_SET,
        init=False,
    )

    def _get_errors(self: SignerListSet) -> Dict[str, str]:
        errors = super()._get_errors()

        # deleting a signer list requires self.signer_quorum == 0 and
        # self.signer_entries is None
        if self.signer_quorum == 0 and self.signer_entries is not None:
            errors["signer_list_set"] = (
                "Must not include a `signer_entries` value if the signer list is being "
                "deleted."
            )
        if self.signer_quorum != 0 and self.signer_entries is None:
            errors["signer_list_set"] = (
                "Must have a value of zero for `signer_quorum` if the signer list is "
                "being deleted."
            )

        if self.signer_entries is None:  # deletion of the SignerList object
            return errors

        if self.signer_quorum <= 0:
            errors[
                "signer_quorum"
            ] = "`signer_quorum` must be greater than or equal to 0."

        if (
            len(self.signer_entries) < 1
            or len(self.signer_entries) > MAX_SIGNER_ENTRIES
        ):
            errors["signer_entries"] = (
                "`signer_entries` must have at least 1 member and no more than {} "
                "members. If this transaction is deleting the SignerList, then "
                "this parameter must be omitted.".format(MAX_SIGNER_ENTRIES)
            )
            return errors

        account_set = set()
        signer_weight_sum = 0

        for signer_entry in self.signer_entries:
            if signer_entry.account == self.account:
                errors["signer_entries"] = (
                    "The account submitting the transaction cannot appear in a "
                    "signer entry."
                )
            if signer_entry.wallet_locator is not None and not bool(
                HEX_WALLET_LOCATOR_REGEX.fullmatch(signer_entry.wallet_locator)
            ):
                errors["signer_entries"] = (
                    "A SignerEntry's wallet_locator must be a 256-bit (32-byte)"
                    "hexadecimal value."
                )
            account_set.add(signer_entry.account)
            signer_weight_sum += signer_entry.signer_weight

        if self.signer_quorum > signer_weight_sum:
            errors["signer_quorum"] = (
                "`signer_quorum` must be less than or equal to the sum of the "
                "SignerWeight values in the `signer_entries` list."
            )

        if len(account_set) != len(self.signer_entries):
            errors["signer_entries"] = (
                "An account cannot appear multiple times in the list of signer "
                "entries."
            )
        return errors



"""The base class for all model types."""

from __future__ import annotations

import json
import re
from abc import ABC
from dataclasses import fields
from enum import Enum
from typing import Any, Dict, List, Pattern, Type, TypeVar, Union, cast, get_type_hints

from typing_extensions import Final, Literal, get_args, get_origin

from xrpl.models.exceptions import XRPLModelException
from xrpl.models.required import REQUIRED
from xrpl.models.types import XRPL_VALUE_TYPE

# this regex splits words based on one of three cases:
#
# 1. 1-or-more non-capital chars at the beginning of the string. Handles cases
# like "value" where the entire string is not capitalized. Would also handle
# true camelCase instead of PascalCase
_CAMEL_CASE_LEADING_LOWER: Final[str] = "^[^A-Z]+"
# 2. 1-or-more capital chars NOT followed by a non-capital char. Handles
# abbreviated PascalCase values like "URI".
_CAMEL_CASE_ABBREVIATION: Final[str] = "[A-Z]+(?![^A-Z])"
# 3. 1 capital char followed by N non-capital-chars. Handles the typical
# PascalCase like "Amount"
_CAMEL_CASE_TYPICAL: Final[str] = "[A-Z][^A-Z]*"
#
# combining the above together into one regex:
_CAMEL_TO_SNAKE_CASE_REGEX: Final[Pattern[str]] = re.compile(
    f"(?:{_CAMEL_CASE_LEADING_LOWER}|{_CAMEL_CASE_ABBREVIATION}|{_CAMEL_CASE_TYPICAL})"
)

# This is used to make exceptions when converting dictionary keys to xrpl JSON
# keys. We snake case keys, but some keys are abbreviations.
ABBREVIATIONS: Final[Dict[str, str]] = {
    "amm": "AMM",
    "id": "ID",
    "lp": "LP",
    "nftoken": "NFToken",
    "unl": "UNL",
    "uri": "URI",
    "xchain": "XChain",
}

BM = TypeVar("BM", bound="BaseModel")  # any type inherited from BaseModel


def _key_to_json(field: str) -> str:
    """
    Transforms camelCase or PascalCase to snake_case. For example:
        1. 'TransactionType' becomes 'transaction_type'
        2. 'value' remains 'value'
        3. 'URI' becomes 'uri'
    """
    # convert all special CamelCase substrings to capitalized strings
    for spec_str in ABBREVIATIONS.values():
        if spec_str in field:
            field = field.replace(spec_str, spec_str.capitalize())

    return "_".join(
        [word.lower() for word in _CAMEL_TO_SNAKE_CASE_REGEX.findall(field)]
    )


def _value_to_json(value: XRPL_VALUE_TYPE) -> XRPL_VALUE_TYPE:
    if isinstance(value, dict):
        return {_key_to_json(k): _value_to_json(v) for (k, v) in value.items()}
    if isinstance(value, list):
        return [_value_to_json(sub_value) for sub_value in value]
    return value


class BaseModel(ABC):
    """The base class for all model types."""

    @classmethod
    def is_dict_of_model(cls: Type[BM], dictionary: Any) -> bool:
        """
        Checks whether the provided ``dictionary`` is a dictionary representation
        of this class.

        **Note:** This only checks the exact model, and does not count model
        inheritance. This method returns ``False`` if the dictionary represents
        a subclass of this class.

        Args:
            dictionary: The dictionary to check.

        Returns:
            True if dictionary is a ``dict`` representation of an instance of this
            class; False if not.
        """
        return (
            isinstance(dictionary, dict)
            and set(get_type_hints(cls).keys()).issuperset(set(dictionary.keys()))
            and all(
                [
                    attr in dictionary
                    for attr, value in get_type_hints(cls).items()
                    if value is REQUIRED
                ]
            )
        )

    @classmethod
    def from_dict(cls: Type[BM], value: Dict[str, XRPL_VALUE_TYPE]) -> BM:
        """
        Construct a new BaseModel from a dictionary of parameters.

        Args:
            value: The value to construct the BaseModel from.

        Returns:
            A new BaseModel object, constructed using the given parameters.

        Raises:
            XRPLModelException: If the dictionary provided is invalid.
        """
        # returns a dictionary mapping class params to their types
        class_types = get_type_hints(cls)

        args = {}
        for param in value:
            if param not in class_types:
                raise XRPLModelException(
                    f"{param} not a valid parameter for {cls.__name__}"
                )

            args[param] = cls._from_dict_single_param(
                param, class_types[param], value[param]
            )

        init = cls._get_only_init_args(args)
        return cls(**init)

    @classmethod
    def _from_dict_single_param(
        cls: Type[BM],
        param: str,
        param_type: Type[Any],
        param_value: Union[int, str, bool, BaseModel, Enum, List[Any], Dict[str, Any]],
    ) -> Any:
        """Recursively handles each individual param in `from_dict`."""
        param_type_origin = get_origin(param_type)
        # returns `list` if a List, `Union` if a Union, None otherwise

        if param_type_origin is list and isinstance(param_value, list):
            # expected a List, received a List
            list_type = get_args(param_type)[0]
            return [
                cls._from_dict_single_param(param, list_type, item)
                for item in param_value
            ]

        if param_type_origin is Union:
            for param_type_option in get_args(param_type):
                # iterate through the types Union-ed together
                try:
                    # try to use this Union-ed type to process param_value
                    return cls._from_dict_single_param(
                        param, param_type_option, param_value
                    )
                except XRPLModelException:
                    # this Union-ed type did not work, move onto the next one
                    pass

        # no more collections (no params expect a Dict)

        if param_type is Any:
            # param_type is Any (e.g. will accept anything)
            return param_value

        if isinstance(param_type, type) and isinstance(param_value, param_type):
            # expected an object, received the correct object
            return param_value

        if get_origin(param_type) == Literal:
            if param_value in get_args(param_type):
                return param_value

        if (
            isinstance(param_type, type)
            and issubclass(param_type, Enum)
            and param_value in list(param_type)
        ):
            # expected an Enum and received a valid value for it.
            # for some reason required for string enums.
            return param_value

        if (
            isinstance(param_type, type)
            and issubclass(param_type, BaseModel)
            and isinstance(param_value, dict)
        ):
            # expected an XRPL Model, received a Dict
            return cast(BaseModel, param_type).from_dict(param_value)

        # received something we didn't expect, raise an error
        if isinstance(param_type, type) and issubclass(param_type, BaseModel):
            error_message = (
                f"{param} expected a {param_type} or a Dict representing {param_type}, "
                f"received a {type(param_value)}"
            )
        else:
            error_message = (
                f"{param} expected a {param_type}, received a {type(param_value)}"
            )
        raise XRPLModelException(error_message)

    @classmethod
    def _get_only_init_args(cls: Type[BM], args: Dict[str, Any]) -> Dict[str, Any]:
        init_keys = {field.name for field in fields(cls) if field.init is True}
        valid_args = {key: value for key, value in args.items() if key in init_keys}
        return valid_args

    @classmethod
    def from_xrpl(cls: Type[BM], value: Union[str, Dict[str, Any]]) -> BM:
        """
        Creates a BaseModel object based on a JSON-like dictionary of keys in the JSON
        format used by the binary codec, or an actual JSON string representing the same
        data.

        Args:
            value: The dictionary or JSON string to be instantiated.

        Returns:
            A BaseModel object instantiated from the input.
        """
        if isinstance(value, str):
            value = json.loads(value)

        formatted_dict = {
            _key_to_json(k): _value_to_json(v)
            for (k, v) in cast(Dict[str, XRPL_VALUE_TYPE], value).items()
        }

        return cls.from_dict(formatted_dict)

    def __post_init__(self: BaseModel) -> None:
        """Called by dataclasses immediately after __init__."""
        self.validate()

    def validate(self: BaseModel) -> None:
        """
        Raises if this object is invalid.

        Raises:
            XRPLModelException: if this object is invalid.
        """
        errors = self._get_errors()
        if len(errors) > 0:
            raise XRPLModelException(str(errors))

    def is_valid(self: BaseModel) -> bool:
        """
        Returns whether this BaseModel is valid.

        Returns:
            Whether this BaseModel is valid.
        """
        return len(self._get_errors()) == 0

    def _get_errors(self: BaseModel) -> Dict[str, str]:
        """
        Extended in subclasses to define custom validation logic.

        Returns:
            Dictionary of any errors found on self.
        """
        return {
            attr: f"{attr} is not set"
            for attr, value in self.__dict__.items()
            if value is REQUIRED
        }

    def to_dict(self: BaseModel) -> Dict[str, Any]:
        """
        Returns the dictionary representation of a BaseModel.

        If not overridden, returns the object dict with all non-None values.

        Returns:
            The dictionary representation of a BaseModel.
        """
        # mypy doesn't realize that BaseModel has a field called __dataclass_fields__
        dataclass_fields = self.__dataclass_fields__.keys()  # type: ignore
        return {
            key: self._to_dict_elem(getattr(self, key))
            for key in dataclass_fields
            if getattr(self, key) is not None
        }

    def _to_dict_elem(self: BaseModel, elem: Any) -> Any:
        if isinstance(elem, BaseModel):
            return elem.to_dict()
        if isinstance(elem, Enum):
            return elem.value
        if isinstance(elem, list):
            return [
                self._to_dict_elem(sub_elem)
                for sub_elem in elem
                if sub_elem is not None
            ]
        return elem

    def __eq__(self: BaseModel, other: object) -> bool:
        """Compares a BaseModel to another object to determine if they are equal."""
        return isinstance(other, BaseModel) and self.to_dict() == other.to_dict()

    def __repr__(self: BaseModel) -> str:
        """Returns a string representation of a BaseModel object"""
        repr_items = [f"{key}={repr(value)}" for key, value in self.to_dict().items()]
        return f"{type(self).__name__}({repr_items})"



"""
Codec for serializing and deserializing a hash field with a width
of 128 bits (16 bytes).
`See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
"""
from __future__ import annotations

from typing import Optional, Type

from xrpl.core.binarycodec.exceptions import XRPLBinaryCodecException
from xrpl.core.binarycodec.types.hash import Hash


class Hash128(Hash):
    """
    Codec for serializing and deserializing a hash field with a width
    of 128 bits (16 bytes).
    `See Hash Fields <https://xrpl.org/serialization.html#hash-fields>`_
    """

    def __init__(self: Hash128, buffer: Optional[bytes]) -> None:
        """
        Construct a Hash128.

        Args:
            buffer: The byte buffer that will be used to store the serialized
                encoding of this field.
        """
        buffer = (
            buffer
            if buffer is not None and len(buffer) > 0
            else bytes(self._get_length())
        )

        if len(buffer) != self._get_length():
            raise XRPLBinaryCodecException(
                f"Invalid hash length {len(buffer)}. Expected {self._get_length()}"
            )
        super().__init__(buffer)

    def __str__(self: Hash128) -> str:
        """Returns a hex-encoded string representation of the bytes buffer."""
        hex = self.to_hex()
        if hex == "0" * len(hex):
            return ""
        return hex

    @classmethod
    def _get_length(cls: Type[Hash128]) -> int:
        return 16



"""Transaction and pseudo-transaction type Enums."""

from xrpl.models.transactions.types.pseudo_transaction_type import PseudoTransactionType
from xrpl.models.transactions.types.transaction_type import TransactionType

__all__ = ["PseudoTransactionType", "TransactionType"]



"""A XChainBridge represents a cross-chain bridge."""

from __future__ import annotations

from dataclasses import dataclass

from xrpl.models.base_model import BaseModel
from xrpl.models.currencies import Currency
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class XChainBridge(BaseModel):
    """A XChainBridge represents a cross-chain bridge."""

    locking_chain_door: str
    locking_chain_issue: Currency
    issuing_chain_door: str
    issuing_chain_issue: Currency



"""Utility functions for the transaction parser."""

from xrpl.utils.txn_parser.utils.balance_parser import derive_account_balances
from xrpl.utils.txn_parser.utils.nodes import NormalizedNode, normalize_nodes
from xrpl.utils.txn_parser.utils.order_book_parser import compute_order_book_changes
from xrpl.utils.txn_parser.utils.parser import get_value
from xrpl.utils.txn_parser.utils.types import AccountBalances, AccountOfferChanges

__all__ = [
    "get_value",
    "derive_account_balances",
    "NormalizedNode",
    "normalize_nodes",
    "AccountBalances",
    "AccountOfferChanges",
    "compute_order_book_changes",
]



"""High-level transaction methods with XRPL transactions."""
import math
from typing import Any, Dict, Optional, cast

from typing_extensions import Final

from xrpl.asyncio.account import get_next_valid_seq_number
from xrpl.asyncio.clients import Client, XRPLRequestFailureException
from xrpl.asyncio.ledger import get_fee, get_latest_validated_ledger_sequence
from xrpl.constants import XRPLException
from xrpl.core.addresscodec import is_valid_xaddress, xaddress_to_classic_address
from xrpl.core.binarycodec import encode, encode_for_multisigning, encode_for_signing
from xrpl.core.keypairs.main import sign as keypairs_sign
from xrpl.models.requests import ServerState, SubmitOnly
from xrpl.models.response import Response
from xrpl.models.transactions import EscrowFinish
from xrpl.models.transactions.transaction import Signer, Transaction
from xrpl.models.transactions.transaction import (
    transaction_json_to_binary_codec_form as model_transaction_to_binary_codec,
)
from xrpl.models.transactions.types.transaction_type import TransactionType
from xrpl.utils import drops_to_xrp, xrp_to_drops
from xrpl.wallet.main import Wallet

_LEDGER_OFFSET: Final[int] = 20

# TODO: make this dynamic based on the current ledger fee
_OWNER_RESERVE_FEE: Final[int] = int(xrp_to_drops(2))


async def sign_and_submit(
    transaction: Transaction,
    wallet: Wallet,
    client: Client,
    autofill: bool = True,
    check_fee: bool = True,
) -> Response:
    """
    Signs a transaction (locally, without trusting external rippled nodes) and submits
    it to the XRPL.

    Args:
        transaction: the transaction to be signed and submitted.
        wallet: the wallet with which to sign the transaction.
        client: the network client with which to submit the transaction.
        autofill: whether to autofill the relevant fields. Defaults to True.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.

    Returns:
        The response from the ledger.
    """
    if autofill:
        transaction = await autofill_and_sign(transaction, wallet, client, check_fee)
    else:
        transaction = await sign(transaction, wallet, check_fee)
    return await submit(transaction, client)


safe_sign_and_submit_transaction = sign_and_submit


async def sign(
    transaction: Transaction,
    wallet: Wallet,
    check_fee: bool = True,
    multisign: bool = False,
) -> Transaction:
    """
    Signs a transaction locally, without trusting external rippled nodes.

    Args:
        transaction: the transaction to be signed.
        wallet: the wallet with which to sign the transaction.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.
        multisign: whether to sign the transaction for a multisignature transaction.

    Returns:
        The signed transaction blob.
    """
    if multisign:
        signature = keypairs_sign(
            bytes.fromhex(
                encode_for_multisigning(
                    transaction.to_xrpl(),
                    wallet.classic_address,
                )
            ),
            wallet.private_key,
        )
        tx_dict = transaction.to_dict()
        tx_dict["signers"] = [
            Signer(
                account=wallet.classic_address,
                txn_signature=signature,
                signing_pub_key=wallet.public_key,
            )
        ]
        return Transaction.from_dict(tx_dict)

    if check_fee:
        await _check_fee(transaction)
    transaction_json = _prepare_transaction(transaction, wallet)
    serialized_for_signing = encode_for_signing(transaction_json)
    serialized_bytes = bytes.fromhex(serialized_for_signing)
    signature = keypairs_sign(serialized_bytes, wallet.private_key)
    transaction_json["TxnSignature"] = signature
    return Transaction.from_xrpl(transaction_json)


safe_sign_transaction = sign


async def autofill_and_sign(
    transaction: Transaction,
    wallet: Wallet,
    client: Client,
    check_fee: bool = True,
) -> Transaction:
    """
    Autofills relevant fields. Then, signs a transaction locally, without trusting
    external rippled nodes.

    Args:
        transaction: the transaction to be signed.
        wallet: the wallet with which to sign the transaction.
        client: a network client.
        check_fee: whether to check if the fee is higher than the expected transaction
            type fee. Defaults to True.

    Returns:
        The signed transaction.
    """
    # We do the transaction fee check here as we have the Client available.
    # The fee check will be done if transaction.fee exists. Otherwise the fee
    # will be auto-filled in autofill()
    if check_fee:
        await _check_fee(transaction, client)

    return await sign(await autofill(transaction, client), wallet, False)


safe_sign_and_autofill_transaction = autofill_and_sign


async def submit(
    transaction: Transaction,
    client: Client,
) -> Response:
    """
    Submits a transaction to the ledger.

    Args:
        transaction: the Transaction to be submitted.
        client: the network client with which to submit the transaction.

    Returns:
        The response from the ledger.

    Raises:
        XRPLRequestFailureException: if the rippled API call fails.
    """
    transaction_blob = encode(transaction.to_xrpl())
    response = await client._request_impl(SubmitOnly(tx_blob=transaction_blob))
    if response.is_successful():
        return response

    raise XRPLRequestFailureException(response.result)


submit_transaction = submit


def _prepare_transaction(
    transaction: Transaction,
    wallet: Wallet,
) -> Dict[str, Any]:
    """
    Prepares a Transaction by converting it to a JSON-like dictionary, converting the
    field names to CamelCase. If a Client is provided, then it also autofills any
    relevant fields.

    Args:
        transaction: the Transaction to be prepared.
        wallet: the wallet that will be used for signing.

    Returns:
        A JSON-like dictionary that is ready to be signed.

    Raises:
        XRPLException: if both LastLedgerSequence and `ledger_offset` are provided, or
            if an address tag is provided that does not match the X-Address tag.
    """
    transaction_json = transaction.to_xrpl()
    transaction_json["SigningPubKey"] = wallet.public_key

    _validate_account_xaddress(transaction_json, "Account", "SourceTag")
    if "Destination" in transaction_json:
        _validate_account_xaddress(transaction_json, "Destination", "DestinationTag")

    # DepositPreauth
    _convert_to_classic_address(transaction_json, "Authorize")
    _convert_to_classic_address(transaction_json, "Unauthorize")
    # EscrowCancel, EscrowFinish
    _convert_to_classic_address(transaction_json, "Owner")
    # SetRegularKey
    _convert_to_classic_address(transaction_json, "RegularKey")

    return transaction_json


async def autofill(
    transaction: Transaction, client: Client, signers_count: Optional[int] = None
) -> Transaction:
    """
    Autofills fields in a transaction. This will set `sequence`, `fee`, and
    `last_ledger_sequence` according to the current state of the server this Client is
    connected to. It also converts all X-Addresses to classic addresses.

    Args:
        transaction: the transaction to be signed.
        client: a network client.
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions.

    Returns:
        The autofilled transaction.
    """
    transaction_json = transaction.to_dict()
    if "sequence" not in transaction_json:
        sequence = await get_next_valid_seq_number(transaction_json["account"], client)
        transaction_json["sequence"] = sequence
    if "fee" not in transaction_json:
        transaction_json["fee"] = await _calculate_fee_per_transaction_type(
            transaction, client, signers_count
        )
    if "last_ledger_sequence" not in transaction_json:
        ledger_sequence = await get_latest_validated_ledger_sequence(client)
        transaction_json["last_ledger_sequence"] = ledger_sequence + _LEDGER_OFFSET
    return Transaction.from_dict(transaction_json)


def _validate_account_xaddress(
    json: Dict[str, Any], account_field: str, tag_field: str
) -> None:
    """
    Mutates JSON-like dictionary so the X-Address in the account field is the classic
    address, and the tag is in the tag field.

    Args:
        json: JSON-like dictionary with transaction data or similar
        account_field: the field of `json` that may contain an X-Address
        tag_field: the field of `json` that may contain a source or destination tag

    Raises:
        XRPLException: if both an X-Address containing a tag and a tag field are
            provided and they do not match.
    """
    if is_valid_xaddress(json[account_field]):
        account, tag, _ = xaddress_to_classic_address(json[account_field])
        json[account_field] = account
        if tag_field in json and json[tag_field] != tag:
            raise XRPLException(f"{tag_field} value does not match X-Address tag")
        json[tag_field] = tag


def _convert_to_classic_address(json: Dict[str, Any], field: str) -> None:
    """
    Mutates JSON-like dictionary to convert the given field from an X-Address (if
    applicable) to a classic address.

    Args:
        json: JSON-like dictionary with transaction data or similar
        field: the field in `json` that may contain an X-Address
    """
    if field in json and is_valid_xaddress(json[field]):
        json[field] = xaddress_to_classic_address(json[field])


def transaction_json_to_binary_codec_form(dictionary: Dict[str, Any]) -> Dict[str, Any]:
    """
    Returns a new dictionary in which the keys have been formatted as CamelCase and
    standardized to be serialized by the binary codec.

    Args:
        dictionary: The dictionary to be reformatted.

    Returns:
        A new dictionary object that has been reformatted.
    """
    return model_transaction_to_binary_codec(dictionary)


async def _check_fee(transaction: Transaction, client: Optional[Client] = None) -> None:
    """
    Checks if the Transaction fee is lower than the expected Transaction type fee.

    Args:
        transaction: The transaction to check.
        client: Client instance to use to look up network load

    Raises:
        XRPLException: if the transaction fee is higher than the expected fee.
    """
    expected_fee = max(
        xrp_to_drops(0.1),  # a fee that is obviously too high
        await _calculate_fee_per_transaction_type(transaction, client),
    )

    if transaction.fee and int(transaction.fee) > int(expected_fee):
        raise XRPLException(
            f"Fee value: {str(drops_to_xrp(transaction.fee))} XRP is likely entered "
            "incorrectly, since it is much larger than the typical XRP transaction "
            "cost. If this is intentional, use `check_fee=False`."
        )


async def _calculate_fee_per_transaction_type(
    transaction: Transaction,
    client: Optional[Client] = None,
    signers_count: Optional[int] = None,
) -> str:
    """
    Calculate the total fee in drops for a transaction based on:
    - the network fee
    - the transaction condition

    https://xrpl.org/transaction-cost.html#special-transaction-costs

    Args:
        transaction: the Transaction to be submitted.
        client: the network client with which to submit the transaction.
        signers_count: the expected number of signers for this transaction.
            Only used for multisigned transactions.

    Returns:
        The expected Transaction fee in drops
    """
    # Reference Transaction (Most transactions)
    if client is None:
        net_fee = 10  # 10 drops
    else:
        net_fee = int(await get_fee(client))  # Usually 0.00001 XRP (10 drops)

    base_fee = net_fee

    # EscrowFinish Transaction with Fulfillment
    # https://xrpl.org/escrowfinish.html#escrowfinish-fields
    if transaction.transaction_type == TransactionType.ESCROW_FINISH:
        escrow_finish = cast(EscrowFinish, transaction)
        if escrow_finish.fulfillment is not None:
            fulfillment_bytes = escrow_finish.fulfillment.encode("ascii")
            # 10 drops × (33 + (Fulfillment size in bytes / 16))
            base_fee = math.ceil(net_fee * (33 + (len(fulfillment_bytes) / 16)))

    # AccountDelete Transaction
    if (
        transaction.transaction_type == TransactionType.ACCOUNT_DELETE
        or transaction.transaction_type == TransactionType.AMM_CREATE
    ):
        if client is None:
            base_fee = _OWNER_RESERVE_FEE
        else:
            base_fee = await _fetch_owner_reserve_fee(client)

    # Multi-signed Transaction
    # 10 drops × (1 + Number of Signatures Provided)
    if signers_count is not None and signers_count > 0:
        base_fee += net_fee * (1 + signers_count)
    # Round Up base_fee and return it as a String
    return str(math.ceil(base_fee))


async def _fetch_owner_reserve_fee(client: Client) -> int:
    server_state = await client._request_impl(ServerState())
    fee = server_state.result["state"]["validated_ledger"]["reserve_inc"]
    return int(fee)



"""
The server_info command asks the server for a
human-readable version of various information
about the rippled server being queried.
"""
from dataclasses import dataclass, field

from xrpl.models.requests.request import Request, RequestMethod
from xrpl.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class ServerInfo(Request):
    """
    The server_info command asks the server for a
    human-readable version of various information
    about the rippled server being queried.
    """

    method: RequestMethod = field(default=RequestMethod.SERVER_INFO, init=False)
